"""Create `flow_run_input` table

Revision ID: 733ca1903976
Revises: 9c493c02ca6d
Create Date: 2023-12-07 09:53:20.009178

"""

import sqlalchemy as sa
from alembic import op

import prefect

# revision identifiers, used by Alembic.
revision = "733ca1903976"
down_revision = "9c493c02ca6d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "flow_run_input",
        sa.Column(
            "id",
            prefect.server.utilities.database.UUID(),
            server_default=sa.text("(GEN_RANDOM_UUID())"),
            nullable=False,
        ),
        sa.Column(
            "created",
            prefect.server.utilities.database.Timestamp(timezone=True),
            server_default=sa.text("CURRENT_TIMESTAMP"),
            nullable=False,
        ),
        sa.Column(
            "updated",
            prefect.server.utilities.database.Timestamp(timezone=True),
            server_default=sa.text("CURRENT_TIMESTAMP"),
            nullable=False,
        ),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("value", sa.Text(), nullable=False),
        sa.Column(
            "flow_run_id", prefect.server.utilities.database.UUID(), nullable=False
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_flow_run_input")),
        sa.UniqueConstraint(
            "flow_run_id", "key", name=op.f("uq_flow_run_input__flow_run_id_key")
        ),
    )
    op.create_index(
        op.f("ix_flow_run_input__updated"), "flow_run_input", ["updated"], unique=False
    )


def downgrade():
    op.drop_index(op.f("ix_flow_run_input__updated"), table_name="flow_run_input")
    op.drop_table("flow_run_input")



"""
The CancellationCleanup service. Responsible for cancelling tasks and subflows that haven't finished.
"""

import asyncio
import datetime
from typing import Any, Optional
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.sql.expression import or_

import prefect.server.models as models
from prefect.server.database import PrefectDBInterface, orm_models
from prefect.server.database.dependencies import db_injector
from prefect.server.schemas import filters, states
from prefect.server.services.base import LoopService
from prefect.settings import PREFECT_API_SERVICES_CANCELLATION_CLEANUP_LOOP_SECONDS
from prefect.settings.context import get_current_settings
from prefect.settings.models.server.services import ServicesBaseSetting
from prefect.types._datetime import now

NON_TERMINAL_STATES = list(set(states.StateType) - states.TERMINAL_STATES)


class CancellationCleanup(LoopService):
    """
    Cancels tasks and subflows of flow runs that have been cancelled
    """

    @classmethod
    def service_settings(cls) -> ServicesBaseSetting:
        return get_current_settings().server.services.cancellation_cleanup

    def __init__(self, loop_seconds: Optional[float] = None, **kwargs: Any):
        super().__init__(
            loop_seconds=loop_seconds
            or PREFECT_API_SERVICES_CANCELLATION_CLEANUP_LOOP_SECONDS.value(),
            **kwargs,
        )

        # query for this many runs to mark failed at once
        self.batch_size = 200

    @db_injector
    async def run_once(self, db: PrefectDBInterface) -> None:
        """
        - cancels active tasks belonging to recently cancelled flow runs
        - cancels any active subflow that belongs to a cancelled flow
        """
        # cancels active tasks belonging to recently cancelled flow runs
        await self.clean_up_cancelled_flow_run_task_runs(db)

        # cancels any active subflow run that belongs to a cancelled flow run
        await self.clean_up_cancelled_subflow_runs(db)

        self.logger.info("Finished cleaning up cancelled flow runs.")

    async def clean_up_cancelled_flow_run_task_runs(
        self, db: PrefectDBInterface
    ) -> None:
        high_water_mark = UUID(int=0)
        while True:
            cancelled_flow_query = (
                sa.select(db.FlowRun)
                .where(
                    db.FlowRun.state_type == states.StateType.CANCELLED,
                    db.FlowRun.end_time.is_not(None),
                    db.FlowRun.end_time >= (now("UTC") - datetime.timedelta(days=1)),
                    db.FlowRun.id > high_water_mark,
                )
                .order_by(db.FlowRun.id)
                .limit(self.batch_size)
            )

            async with db.session_context() as session:
                flow_run_result = await session.execute(cancelled_flow_query)
            flow_runs = flow_run_result.scalars().all()

            for run in flow_runs:
                await self._cancel_child_runs(db=db, flow_run=run)
                high_water_mark = run.id

            # if no relevant flows were found, exit the loop
            if len(flow_runs) < self.batch_size:
                break

    async def clean_up_cancelled_subflow_runs(self, db: PrefectDBInterface) -> None:
        high_water_mark = UUID(int=0)
        while True:
            # Performance optimization: Load only required columns while maintaining ORM functionality
            # Required columns:
            # - id: for high water mark tracking
            # - state_type: for state validation
            # - parent_task_run_id: for parent task run checks
            # - deployment_id: for determining cancellation state type
            subflow_query = (
                sa.select(db.FlowRun)
                .options(
                    sa.orm.load_only(
                        db.FlowRun.id,
                        db.FlowRun.state_type,
                        db.FlowRun.parent_task_run_id,
                        db.FlowRun.deployment_id,
                    ),
                )
                .where(
                    or_(
                        db.FlowRun.state_type == states.StateType.PENDING,
                        db.FlowRun.state_type == states.StateType.SCHEDULED,
                        db.FlowRun.state_type == states.StateType.RUNNING,
                        db.FlowRun.state_type == states.StateType.PAUSED,
                        db.FlowRun.state_type == states.StateType.CANCELLING,
                    ),
                    db.FlowRun.id > high_water_mark,
                    db.FlowRun.parent_task_run_id.is_not(None),
                )
                .order_by(db.FlowRun.id)
                .limit(self.batch_size)
            )

            async with db.session_context() as session:
                subflow_run_result = await session.execute(subflow_query)
            subflow_runs = subflow_run_result.scalars().all()

            for subflow_run in subflow_runs:
                await self._cancel_subflow(db=db, flow_run=subflow_run)
                high_water_mark = max(high_water_mark, subflow_run.id)

            # if no relevant flows were found, exit the loop
            if len(subflow_runs) < self.batch_size:
                break

    async def _cancel_child_runs(
        self, db: PrefectDBInterface, flow_run: orm_models.FlowRun
    ) -> None:
        async with db.session_context() as session:
            child_task_runs = await models.task_runs.read_task_runs(
                session,
                flow_run_filter=filters.FlowRunFilter(
                    id=filters.FlowRunFilterId(any_=[flow_run.id])
                ),
                task_run_filter=filters.TaskRunFilter(
                    state=filters.TaskRunFilterState(
                        type=filters.TaskRunFilterStateType(any_=NON_TERMINAL_STATES)
                    )
                ),
                limit=100,
            )

        for task_run in child_task_runs:
            async with db.session_context(begin_transaction=True) as session:
                await models.task_runs.set_task_run_state(
                    session=session,
                    task_run_id=task_run.id,
                    state=states.Cancelled(
                        message="The parent flow run was cancelled."
                    ),
                    force=True,
                )

    async def _cancel_subflow(
        self, db: PrefectDBInterface, flow_run: orm_models.FlowRun
    ) -> Optional[bool]:
        if not flow_run.parent_task_run_id or not flow_run.state:
            return False

        if flow_run.state.type in states.TERMINAL_STATES:
            return False

        async with db.session_context() as session:
            parent_task_run = await models.task_runs.read_task_run(
                session, task_run_id=flow_run.parent_task_run_id
            )

            if not parent_task_run or not parent_task_run.flow_run_id:
                # Global orchestration policy will prevent further orchestration
                return False

            containing_flow_run = await models.flow_runs.read_flow_run(
                session, flow_run_id=parent_task_run.flow_run_id
            )

            if (
                containing_flow_run
                and containing_flow_run.state
                and containing_flow_run.state.type != states.StateType.CANCELLED
            ):
                # Nothing to do here; the parent is not cancelled
                return False

            if flow_run.deployment_id:
                state = states.Cancelling(message="The parent flow run was cancelled.")
            else:
                state = states.Cancelled(message="The parent flow run was cancelled.")

        async with db.session_context(begin_transaction=True) as session:
            await models.flow_runs.set_flow_run_state(
                session=session,
                flow_run_id=flow_run.id,
                state=state,
            )


if __name__ == "__main__":
    asyncio.run(CancellationCleanup(handle_signals=True).start())



"""
Utilities for querying flow and task run history.
"""

import datetime
import json
from typing import TYPE_CHECKING, List, Optional

import pydantic
import sqlalchemy as sa
from typing_extensions import Literal

import prefect.server.models as models
import prefect.server.schemas as schemas
from prefect.logging import get_logger
from prefect.server.database import PrefectDBInterface, db_injector
from prefect.types import DateTime

if TYPE_CHECKING:
    import logging

logger: "logging.Logger" = get_logger("server.api")


@db_injector
async def run_history(
    db: PrefectDBInterface,
    session: sa.orm.Session,
    run_type: Literal["flow_run", "task_run"],
    history_start: DateTime,
    history_end: DateTime,
    history_interval: datetime.timedelta,
    flows: Optional[schemas.filters.FlowFilter] = None,
    flow_runs: Optional[schemas.filters.FlowRunFilter] = None,
    task_runs: Optional[schemas.filters.TaskRunFilter] = None,
    deployments: Optional[schemas.filters.DeploymentFilter] = None,
    work_pools: Optional[schemas.filters.WorkPoolFilter] = None,
    work_queues: Optional[schemas.filters.WorkQueueFilter] = None,
) -> list[schemas.responses.HistoryResponse]:
    """
    Produce a history of runs aggregated by interval and state
    """

    # SQLite has issues with very small intervals
    # (by 0.001 seconds it stops incrementing the interval)
    if history_interval < datetime.timedelta(seconds=1):
        raise ValueError("History interval must not be less than 1 second.")

    # prepare run-specific models
    if run_type == "flow_run":
        run_model = db.FlowRun
        run_filter_function = models.flow_runs._apply_flow_run_filters
    elif run_type == "task_run":
        run_model = db.TaskRun
        run_filter_function = models.task_runs._apply_task_run_filters
    else:
        raise ValueError(
            f"Unknown run type {run_type!r}. Expected 'flow_run' or 'task_run'."
        )

    # create a CTE for timestamp intervals
    intervals = db.queries.make_timestamp_intervals(
        history_start,
        history_end,
        history_interval,
    ).cte("intervals")

    # apply filters to the flow runs (and related states)
    runs = (
        await run_filter_function(
            db,
            sa.select(
                run_model.id,
                run_model.expected_start_time,
                run_model.estimated_run_time,
                run_model.estimated_start_time_delta,
                run_model.state_type,
                run_model.state_name,
            ).select_from(run_model),
            flow_filter=flows,
            flow_run_filter=flow_runs,
            task_run_filter=task_runs,
            deployment_filter=deployments,
            work_pool_filter=work_pools,
            work_queue_filter=work_queues,
        )
    ).alias("runs")
    # outer join intervals to the filtered runs to create a dataset composed of
    # every interval and the aggregate of all its runs. The runs aggregate is represented
    # by a descriptive JSON object
    counts = (
        sa.select(
            intervals.c.interval_start,
            intervals.c.interval_end,
            # build a JSON object, ignoring the case where the count of runs is 0
            sa.case(
                (sa.func.count(runs.c.id) == 0, None),
                else_=db.queries.build_json_object(
                    "state_type",
                    runs.c.state_type,
                    "state_name",
                    runs.c.state_name,
                    "count_runs",
                    sa.func.count(runs.c.id),
                    # estimated run times only includes positive run times (to avoid any unexpected corner cases)
                    "sum_estimated_run_time",
                    sa.func.sum(
                        sa.func.greatest(
                            0, sa.extract("epoch", runs.c.estimated_run_time)
                        )
                    ),
                    # estimated lateness is the sum of any positive start time deltas
                    "sum_estimated_lateness",
                    sa.func.sum(
                        sa.func.greatest(
                            0, sa.extract("epoch", runs.c.estimated_start_time_delta)
                        )
                    ),
                ),
            ).label("state_agg"),
        )
        .select_from(intervals)
        .join(
            runs,
            sa.and_(
                runs.c.expected_start_time >= intervals.c.interval_start,
                runs.c.expected_start_time < intervals.c.interval_end,
            ),
            isouter=True,
        )
        .group_by(
            intervals.c.interval_start,
            intervals.c.interval_end,
            runs.c.state_type,
            runs.c.state_name,
        )
    ).alias("counts")

    # aggregate all state-aggregate objects into a single array for each interval,
    # ensuring that intervals with no runs have an empty array
    query = (
        sa.select(
            counts.c.interval_start,
            counts.c.interval_end,
            sa.func.coalesce(
                db.queries.json_arr_agg(
                    db.queries.cast_to_json(counts.c.state_agg)
                ).filter(counts.c.state_agg.is_not(None)),
                sa.literal("[]", literal_execute=True),
            ).label("states"),
        )
        .group_by(counts.c.interval_start, counts.c.interval_end)
        .order_by(counts.c.interval_start)
        # return no more than 500 bars
        .limit(500)
    )

    # issue the query
    result = await session.execute(query)
    records = result.mappings()

    # load and parse the record if the database returns JSON as strings
    if db.queries.uses_json_strings:
        records = [dict(r) for r in records]
        for r in records:
            r["states"] = json.loads(r["states"])

    return pydantic.TypeAdapter(
        List[schemas.responses.HistoryResponse]
    ).validate_python(records)



import warnings
from operator import itemgetter
from typing import Any, cast

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema, to_json
from typing_extensions import Self, TypeVar

T = TypeVar("T", infer_variance=True)


class BaseAnnotation(tuple[T]):
    """
    Base class for Prefect annotation types.

    Inherits from `tuple` for unpacking support in other tools.
    """

    __slots__ = ()

    def __new__(cls, value: T) -> Self:
        return super().__new__(cls, (value,))

    # use itemgetter to minimise overhead, just like namedtuple generated code would
    value: T = cast(T, property(itemgetter(0)))

    def unwrap(self) -> T:
        return self[0]

    def rewrap(self, value: T) -> Self:
        return type(self)(value)

    def __eq__(self, other: Any) -> bool:
        if type(self) is not type(other):
            return False
        return super().__eq__(other)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self[0]!r})"


class unmapped(BaseAnnotation[T]):
    """
    Wrapper for iterables.

    Indicates that this input should be sent as-is to all runs created during a mapping
    operation instead of being split.
    """

    def __getitem__(self, _: object) -> T:  # type: ignore[override]  # pyright: ignore[reportIncompatibleMethodOverride]
        # Internally, this acts as an infinite array where all items are the same value
        return super().__getitem__(0)


class allow_failure(BaseAnnotation[T]):
    """
    Wrapper for states or futures.

    Indicates that the upstream run for this input can be failed.

    Generally, Prefect will not allow a downstream run to start if any of its inputs
    are failed. This annotation allows you to opt into receiving a failed input
    downstream.

    If the input is from a failed run, the attached exception will be passed to your
    function.
    """


class quote(BaseAnnotation[T]):
    """
    Simple wrapper to mark an expression as a different type so it will not be coerced
    by Prefect. For example, if you want to return a state from a flow without having
    the flow assume that state.

    quote will also instruct prefect to ignore introspection of the wrapped object
    when passed as flow or task parameter. Parameter introspection can be a
    significant performance hit when the object is a large collection,
    e.g. a large dictionary or DataFrame, and each element needs to be visited. This
    will disable task dependency tracking for the wrapped object, but likely will
    increase performance.

    ```
    @task
    def my_task(df):
        ...

    @flow
    def my_flow():
        my_task(quote(df))
    ```
    """

    def unquote(self) -> T:
        return self.unwrap()


# Backwards compatibility stub for `Quote` class
class Quote(quote[T]):
    def __new__(cls, expr: T) -> Self:
        warnings.warn(
            "Use of `Quote` is deprecated. Use `quote` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return super().__new__(cls, expr)


class NotSet:
    """
    Singleton to distinguish `None` from a value that is not provided by the user.
    """


class freeze(BaseAnnotation[T]):
    """
    Wrapper for parameters in deployments.

    Indicates that this parameter should be frozen in the UI and not editable
    when creating flow runs from this deployment.

    Example:
    ```python
    @flow
    def my_flow(customer_id: str):
        # flow logic

    deployment = my_flow.deploy(parameters={"customer_id": freeze("customer123")})
    ```
    """

    def __new__(cls, value: T) -> Self:
        try:
            to_json(value)
        except Exception:
            raise ValueError("Value must be JSON serializable")
        return super().__new__(cls, value)

    def unfreeze(self) -> T:
        """Return the unwrapped value."""
        return self.unwrap()

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls,  # Use the class itself as the validator
            core_schema.any_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda x: x.unfreeze()  # Serialize by unwrapping the value
            ),
        )



"""
Implementations of `Waiter`s, which allow work to be sent back to a thread while it
waits for the result of the call.
"""

import abc
import asyncio
import contextlib
import inspect
import queue
import threading
from collections import deque
from collections.abc import AsyncGenerator, Awaitable, Generator
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar
from weakref import WeakKeyDictionary

import anyio

from prefect._internal.concurrency import logger
from prefect._internal.concurrency.calls import Call, Portal
from prefect._internal.concurrency.event_loop import call_soon_in_loop
from prefect._internal.concurrency.primitives import Event

T = TypeVar("T")


# Waiters are stored in a queue for each thread
_WAITERS_BY_THREAD: "WeakKeyDictionary[threading.Thread, deque[Waiter[Any]]]" = (
    WeakKeyDictionary()
)


def add_waiter_for_thread(waiter: "Waiter[Any]", thread: threading.Thread) -> None:
    """
    Add a waiter for a thread.
    """
    if thread not in _WAITERS_BY_THREAD:
        _WAITERS_BY_THREAD[thread] = deque()

    _WAITERS_BY_THREAD[thread].append(waiter)


class Waiter(Portal, abc.ABC, Generic[T]):
    """
    A waiter allows waiting for a call while routing callbacks to the
    the current thread.

    Calls sent back to the waiter will be executed when waiting for the result.
    """

    def __init__(self, call: Call[T]) -> None:
        if not TYPE_CHECKING:
            if not isinstance(call, Call):  # Guard against common mistake
                raise TypeError(f"Expected call of type `Call`; got {call!r}.")

        self._call = call
        self._owner_thread = threading.current_thread()

        # Set the waiter for the current thread
        add_waiter_for_thread(self, self._owner_thread)
        super().__init__()

    def call_is_done(self) -> bool:
        return self._call.future.done()

    @abc.abstractmethod
    def wait(self) -> T:
        """
        Wait for the call to finish.

        Watch for and execute any waiting callbacks.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def add_done_callback(self, callback: Call[Any]) -> None:
        """
        Schedule a call to run when the waiter is done waiting.

        If the waiter is already done, a `RuntimeError` error will be thrown.
        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} call={self._call},"
            f" owner={self._owner_thread.name!r}>"
        )


class SyncWaiter(Waiter[T]):
    # Implementation of `Waiter` for use in synchronous contexts

    def __init__(self, call: Call[T]) -> None:
        super().__init__(call=call)
        self._queue: queue.Queue[Optional[Call[T]]] = queue.Queue()
        self._done_callbacks: list[Call[Any]] = []
        self._done_event = threading.Event()

    def submit(self, call: Call[T]) -> Call[T]:
        """
        Submit a callback to execute while waiting.
        """
        if self.call_is_done():
            raise RuntimeError(f"The call {self._call} is already done.")

        self._queue.put_nowait(call)
        call.set_runner(self)
        return call

    def _handle_waiting_callbacks(self) -> None:
        logger.debug("Waiter %r watching for callbacks", self)
        while True:
            callback = self._queue.get()
            if callback is None:
                break

            # Ensure that callbacks are cancelled if the parent call is cancelled so
            # waiting never runs longer than the call
            self._call.future.add_cancel_callback(callback.future.cancel)
            callback.run()
            del callback

    @contextlib.contextmanager
    def _handle_done_callbacks(self) -> Generator[None, Any, None]:
        try:
            yield
        finally:
            # Call done callbacks
            while self._done_callbacks:
                callback = self._done_callbacks.pop()
                if callback:
                    callback.run()

    def add_done_callback(self, callback: Call[Any]) -> None:
        if self._done_event.is_set():
            raise RuntimeError("Cannot add done callbacks to done waiters.")
        else:
            self._done_callbacks.append(callback)

    def wait(self) -> Call[T]:
        # Stop watching for work once the future is done
        self._call.future.add_done_callback(lambda _: self._queue.put_nowait(None))
        self._call.future.add_done_callback(lambda _: self._done_event.set())

        with self._handle_done_callbacks():
            self._handle_waiting_callbacks()

            # Wait for the future to be done
            self._done_event.wait()

        _WAITERS_BY_THREAD[self._owner_thread].remove(self)
        return self._call


class AsyncWaiter(Waiter[T]):
    # Implementation of `Waiter` for use in asynchronous contexts

    def __init__(self, call: Call[T]) -> None:
        super().__init__(call=call)

        # Delay instantiating loop and queue as there may not be a loop present yet
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._queue: Optional[asyncio.Queue[Optional[Call[T]]]] = None
        self._early_submissions: list[Call[T]] = []
        self._done_callbacks: list[Call[Any]] = []
        self._done_event = Event()
        self._done_waiting = False

    def submit(self, call: Call[T]) -> Call[T]:
        """
        Submit a callback to execute while waiting.
        """
        if self.call_is_done():
            raise RuntimeError(f"The call {self._call} is already done.")

        call.set_runner(self)

        if not self._queue:
            # If the loop is not yet available, just push the call to a stack
            self._early_submissions.append(call)
            return call

        # We must put items in the queue from the event loop that owns it
        if TYPE_CHECKING:
            assert self._loop is not None
        call_soon_in_loop(self._loop, self._queue.put_nowait, call)
        return call

    def _resubmit_early_submissions(self) -> None:
        if TYPE_CHECKING:
            assert self._queue is not None
            assert self._loop is not None
        for call in self._early_submissions:
            # We must put items in the queue from the event loop that owns it
            call_soon_in_loop(self._loop, self._queue.put_nowait, call)
        self._early_submissions = []

    async def _handle_waiting_callbacks(self) -> None:
        logger.debug("Waiter %r watching for callbacks", self)
        tasks: list[Awaitable[None]] = []

        if TYPE_CHECKING:
            assert self._queue is not None

        try:
            while True:
                callback = await self._queue.get()
                if callback is None:
                    break

                # Ensure that callbacks are cancelled if the parent call is cancelled so
                # waiting never runs longer than the call
                self._call.future.add_cancel_callback(callback.future.cancel)
                retval = callback.run()
                if inspect.isawaitable(retval):
                    tasks.append(retval)

                del callback

            # Tasks are collected and awaited as a group; if each task was awaited in
            # the above loop, async work would not be executed concurrently
            await asyncio.gather(*tasks)
        finally:
            self._done_waiting = True

    @contextlib.asynccontextmanager
    async def _handle_done_callbacks(self) -> AsyncGenerator[None, Any]:
        try:
            yield
        finally:
            # Call done callbacks
            while self._done_callbacks:
                callback = self._done_callbacks.pop()
                if callback:
                    # We shield against cancellation so we can run the callback
                    with anyio.CancelScope(shield=True):
                        await self._run_done_callback(callback)

    async def _run_done_callback(self, callback: Call[Any]) -> None:
        coro = callback.run()
        if coro:
            await coro

    def add_done_callback(self, callback: Call[Any]) -> None:
        if self._done_event.is_set():
            raise RuntimeError("Cannot add done callbacks to done waiters.")
        else:
            self._done_callbacks.append(callback)

    def _signal_stop_waiting(self) -> None:
        # Only send a `None` to the queue if the waiter is still blocked reading from
        # the queue. Otherwise, it's possible that the event loop is stopped.
        if not self._done_waiting:
            assert self._loop is not None
            assert self._queue is not None
            call_soon_in_loop(self._loop, self._queue.put_nowait, None)

    async def wait(self) -> Call[T]:
        # Assign the loop
        self._loop = asyncio.get_running_loop()
        self._queue = asyncio.Queue()
        self._resubmit_early_submissions()

        # Stop watching for work once the future is done
        self._call.future.add_done_callback(lambda _: self._signal_stop_waiting())
        self._call.future.add_done_callback(lambda _: self._done_event.set())

        async with self._handle_done_callbacks():
            await self._handle_waiting_callbacks()

            # Wait for the future to be done
            await self._done_event.wait()

        _WAITERS_BY_THREAD[self._owner_thread].remove(self)
        return self._call



from datetime import timedelta
from typing import ClassVar

from pydantic import AliasChoices, AliasPath, Field
from pydantic_settings import SettingsConfigDict

from prefect.settings.base import PrefectBaseSettings, build_settings_config


class ServerEventsSettings(PrefectBaseSettings):
    """
    Settings for controlling behavior of the events subsystem
    """

    model_config: ClassVar[SettingsConfigDict] = build_settings_config(
        ("server", "events")
    )

    ###########################################################################
    # Events settings

    stream_out_enabled: bool = Field(
        default=True,
        description="Whether or not to stream events out to the API via websockets.",
        validation_alias=AliasChoices(
            AliasPath("stream_out_enabled"),
            "prefect_server_events_stream_out_enabled",
            "prefect_api_events_stream_out_enabled",
        ),
    )

    related_resource_cache_ttl: timedelta = Field(
        default=timedelta(minutes=5),
        description="The number of seconds to cache related resources for in the API.",
        validation_alias=AliasChoices(
            AliasPath("related_resource_cache_ttl"),
            "prefect_server_events_related_resource_cache_ttl",
            "prefect_api_events_related_resource_cache_ttl",
        ),
    )

    maximum_labels_per_resource: int = Field(
        default=500,
        description="The maximum number of labels a resource may have.",
        validation_alias=AliasChoices(
            AliasPath("maximum_labels_per_resource"),
            "prefect_server_events_maximum_labels_per_resource",
            "prefect_events_maximum_labels_per_resource",
        ),
    )

    maximum_related_resources: int = Field(
        default=100,
        description="The maximum number of related resources an Event may have.",
        validation_alias=AliasChoices(
            AliasPath("maximum_related_resources"),
            "prefect_server_events_maximum_related_resources",
            "prefect_events_maximum_related_resources",
        ),
    )

    maximum_size_bytes: int = Field(
        default=1_500_000,
        description="The maximum size of an Event when serialized to JSON",
        validation_alias=AliasChoices(
            AliasPath("maximum_size_bytes"),
            "prefect_server_events_maximum_size_bytes",
            "prefect_events_maximum_size_bytes",
        ),
    )

    expired_bucket_buffer: timedelta = Field(
        default=timedelta(seconds=60),
        description="The amount of time to retain expired automation buckets",
        validation_alias=AliasChoices(
            AliasPath("expired_bucket_buffer"),
            "prefect_server_events_expired_bucket_buffer",
            "prefect_events_expired_bucket_buffer",
        ),
    )

    proactive_granularity: timedelta = Field(
        default=timedelta(seconds=5),
        description="How frequently proactive automations are evaluated",
        validation_alias=AliasChoices(
            AliasPath("proactive_granularity"),
            "prefect_server_events_proactive_granularity",
            "prefect_events_proactive_granularity",
        ),
    )

    retention_period: timedelta = Field(
        default=timedelta(days=7),
        description="The amount of time to retain events in the database.",
        validation_alias=AliasChoices(
            AliasPath("retention_period"),
            "prefect_server_events_retention_period",
            "prefect_events_retention_period",
        ),
    )

    maximum_websocket_backfill: timedelta = Field(
        default=timedelta(minutes=15),
        description="The maximum range to look back for backfilling events for a websocket subscriber.",
        validation_alias=AliasChoices(
            AliasPath("maximum_websocket_backfill"),
            "prefect_server_events_maximum_websocket_backfill",
            "prefect_events_maximum_websocket_backfill",
        ),
    )

    websocket_backfill_page_size: int = Field(
        default=250,
        gt=0,
        description="The page size for the queries to backfill events for websocket subscribers.",
        validation_alias=AliasChoices(
            AliasPath("websocket_backfill_page_size"),
            "prefect_server_events_websocket_backfill_page_size",
            "prefect_events_websocket_backfill_page_size",
        ),
    )

    messaging_broker: str = Field(
        default="prefect.server.utilities.messaging.memory",
        description="Which message broker implementation to use for the messaging system, should point to a module that exports a Publisher and Consumer class.",
        validation_alias=AliasChoices(
            AliasPath("messaging_broker"),
            "prefect_server_events_messaging_broker",
            "prefect_messaging_broker",
        ),
    )

    messaging_cache: str = Field(
        default="prefect.server.utilities.messaging.memory",
        description="Which cache implementation to use for the events system. Should point to a module that exports a Cache class.",
        validation_alias=AliasChoices(
            AliasPath("messaging_cache"),
            "prefect_server_events_messaging_cache",
            "prefect_messaging_cache",
        ),
    )

    maximum_event_name_length: int = Field(
        default=1024,
        gt=0,
        description="The maximum length of an event name.",
        validation_alias=AliasChoices(
            AliasPath("maximum_event_name_length"),
            "prefect_server_events_maximum_event_name_length",
        ),
    )



from __future__ import annotations

import datetime
import sys
from contextlib import contextmanager
from typing import Any, cast
from unittest import mock
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError, available_timezones

import humanize
from dateutil.parser import parse
from typing_extensions import TypeAlias

if sys.version_info >= (3, 13):
    DateTime: TypeAlias = datetime.datetime
    Date: TypeAlias = datetime.date
    Duration: TypeAlias = datetime.timedelta
else:
    import pendulum
    import pendulum.tz
    from pydantic_extra_types.pendulum_dt import (
        Date as PydanticDate,
    )
    from pydantic_extra_types.pendulum_dt import (
        DateTime as PydanticDateTime,
    )
    from pydantic_extra_types.pendulum_dt import (
        Duration as PydanticDuration,
    )

    DateTime: TypeAlias = PydanticDateTime
    Date: TypeAlias = PydanticDate
    Duration: TypeAlias = PydanticDuration


def parse_datetime(dt: str) -> datetime.datetime:
    if sys.version_info >= (3, 13):
        parsed_dt = parse(dt)
        if parsed_dt.tzinfo is None:
            # Assume UTC if no timezone is provided
            return parsed_dt.replace(tzinfo=ZoneInfo("UTC"))
        else:
            return parsed_dt
    else:
        return cast(datetime.datetime, pendulum.parse(dt))


def get_timezones() -> tuple[str, ...]:
    return tuple(available_timezones())


def create_datetime_instance(v: datetime.datetime) -> datetime.datetime:
    if sys.version_info >= (3, 13):
        if v.tzinfo is None:
            # Assume UTC if no timezone is provided
            return v.replace(tzinfo=ZoneInfo("UTC"))
        else:
            return v

    return DateTime.instance(v)


def from_timestamp(ts: float, tz: str | Any = "UTC") -> datetime.datetime:
    if sys.version_info >= (3, 13):
        if not isinstance(tz, str):
            # Handle pendulum edge case
            tz = tz.name
        return datetime.datetime.fromtimestamp(ts, ZoneInfo(tz))

    return pendulum.from_timestamp(ts, tz)


def human_friendly_diff(
    dt: datetime.datetime | None, other: datetime.datetime | None = None
) -> str:
    if dt is None:
        return ""

    def _normalize(ts: datetime.datetime) -> datetime.datetime:
        """Return *ts* with a valid ZoneInfo; fall back to UTC if needed."""
        if ts.tzinfo is None:
            local_tz = datetime.datetime.now().astimezone().tzinfo
            return ts.replace(tzinfo=local_tz).astimezone(ZoneInfo("UTC"))

        if isinstance(ts.tzinfo, ZoneInfo):
            return ts  # already valid

        if tz_name := getattr(ts.tzinfo, "name", None):
            try:
                return ts.replace(tzinfo=ZoneInfo(tz_name))
            except ZoneInfoNotFoundError:
                pass

        return ts.astimezone(ZoneInfo("UTC"))

    dt = _normalize(dt)

    if other is not None:
        other = _normalize(other)

    if sys.version_info >= (3, 13):
        # humanize expects ZoneInfo or None
        return humanize.naturaltime(dt, when=other)

    # Ensure consistency for pendulum path by using UTC
    pendulum_dt = DateTime.instance(dt.astimezone(ZoneInfo("UTC")))
    pendulum_other = (
        DateTime.instance(other.astimezone(ZoneInfo("UTC"))) if other else None
    )
    return pendulum_dt.diff_for_humans(other=pendulum_other)


def now(
    tz: str | Any = "UTC",
) -> datetime.datetime:
    if sys.version_info >= (3, 13):
        from whenever import ZonedDateTime

        if isinstance(getattr(tz, "name", None), str):
            tz = tz.name

        return ZonedDateTime.now(tz).py_datetime()
    else:
        return pendulum.now(tz)


def end_of_period(dt: datetime.datetime, period: str) -> datetime.datetime:
    """
    Returns the end of the specified unit of time.

    Args:
        dt: The datetime to get the end of.
        period: The period to get the end of.
                Valid values: 'second', 'minute', 'hour', 'day',
                'week'

    Returns:
        DateTime: A new DateTime representing the end of the specified unit.

    Raises:
        ValueError: If an invalid unit is specified.
    """
    if sys.version_info >= (3, 13):
        from whenever import Weekday, ZonedDateTime, days

        if not isinstance(dt.tzinfo, ZoneInfo):
            zdt = ZonedDateTime.from_py_datetime(
                dt.replace(tzinfo=ZoneInfo(dt.tzname() or "UTC"))
            )
        else:
            zdt = ZonedDateTime.from_py_datetime(dt)
        if period == "second":
            zdt = zdt.replace(nanosecond=999999999)
        elif period == "minute":
            zdt = zdt.replace(second=59, nanosecond=999999999)
        elif period == "hour":
            zdt = zdt.replace(minute=59, second=59, nanosecond=999999999)
        elif period == "day":
            zdt = zdt.replace(hour=23, minute=59, second=59, nanosecond=999999999)
        elif period == "week":
            days_till_end_of_week: int = (
                Weekday.SUNDAY.value - zdt.date().day_of_week().value
            )
            zdt = zdt + days(days_till_end_of_week)
            zdt = zdt.replace(
                hour=23,
                minute=59,
                second=59,
                nanosecond=999999999,
            )
        else:
            raise ValueError(f"Invalid period: {period}")

        return zdt.py_datetime()
    else:
        return DateTime.instance(dt).end_of(period)


def start_of_day(dt: datetime.datetime | DateTime) -> datetime.datetime:
    """
    Returns the start of the specified unit of time.

    Args:
        dt: The datetime to get the start of.

    Returns:
        datetime.datetime: A new datetime.datetime representing the start of the specified unit.

    Raises:
        ValueError: If an invalid unit is specified.
    """
    if sys.version_info >= (3, 13):
        from whenever import ZonedDateTime

        if hasattr(dt, "tz"):
            zdt = ZonedDateTime.from_timestamp(
                dt.timestamp(), tz=dt.tz.name if dt.tz else "UTC"
            )
        else:
            zdt = ZonedDateTime.from_py_datetime(dt)

        return zdt.start_of_day().py_datetime()
    else:
        return DateTime.instance(dt).start_of("day")


def earliest_possible_datetime() -> datetime.datetime:
    return datetime.datetime.min.replace(tzinfo=ZoneInfo("UTC"))


@contextmanager
def travel_to(dt: Any):
    if sys.version_info >= (3, 13):
        with mock.patch("prefect.types._datetime.now", return_value=dt):
            yield

    else:
        from pendulum import travel_to

        with travel_to(dt, freeze=True):
            yield


def in_local_tz(dt: datetime.datetime) -> datetime.datetime:
    if sys.version_info >= (3, 13):
        from whenever import LocalDateTime, ZonedDateTime

        if dt.tzinfo is None:
            wdt = LocalDateTime.from_py_datetime(dt)
        else:
            if not isinstance(dt.tzinfo, ZoneInfo):
                if key := getattr(dt.tzinfo, "key", None):
                    dt = dt.replace(tzinfo=ZoneInfo(key))
                else:
                    utc_dt = dt.astimezone(datetime.timezone.utc)
                    dt = utc_dt.replace(tzinfo=ZoneInfo("UTC"))

            wdt = ZonedDateTime.from_py_datetime(dt).to_system_tz()

        return wdt.py_datetime()

    return DateTime.instance(dt).in_tz(pendulum.tz.local_timezone())


def to_datetime_string(dt: datetime.datetime, include_tz: bool = True) -> str:
    if include_tz:
        return dt.strftime("%Y-%m-%d %H:%M:%S %Z")
    else:
        return dt.strftime("%Y-%m-%d %H:%M:%S")



import asyncio
from collections.abc import Generator
from contextlib import contextmanager
from typing import Optional, TypeVar, Union
from uuid import UUID

from prefect.types._datetime import now

from ._asyncio import acquire_concurrency_slots, release_concurrency_slots
from ._events import (
    emit_concurrency_acquisition_events,
    emit_concurrency_release_events,
)

T = TypeVar("T")


@contextmanager
def concurrency(
    names: Union[str, list[str]],
    task_run_id: UUID,
    timeout_seconds: Optional[float] = None,
) -> Generator[None, None, None]:
    """
    A context manager that acquires and releases concurrency slots from the
    given concurrency limits.

    Args:
        names: The names of the concurrency limits to acquire.
        task_run_id: The task run ID acquiring the limits.
        timeout_seconds: The number of seconds to wait to acquire the limits before
            raising a `TimeoutError`. A timeout of `None` will wait indefinitely.

    Raises:
        TimeoutError: If the limits are not acquired within the given timeout.

    Example:
    A simple example of using the sync `concurrency` context manager:
    ```python
    from prefect.concurrency.v1.sync import concurrency

    def resource_heavy():
        with concurrency("test"):
            print("Resource heavy task")

    def main():
        resource_heavy()
    ```
    """
    if not names:
        yield
        return

    names = names if isinstance(names, list) else [names]

    force = {"_sync": True}
    result = acquire_concurrency_slots(
        names, timeout_seconds=timeout_seconds, task_run_id=task_run_id, **force
    )
    assert not asyncio.iscoroutine(result)
    limits = result
    acquisition_time = now("UTC")
    emitted_events = emit_concurrency_acquisition_events(limits, task_run_id)

    try:
        yield
    finally:
        occupancy_period = now("UTC") - acquisition_time
        release_concurrency_slots(
            names, task_run_id, occupancy_period.total_seconds(), **force
        )
        emit_concurrency_release_events(limits, emitted_events, task_run_id)



"""Add ix_events__event_related_occurred_md5

Revision ID: 7a73514ca2d6
Revises: 06b7c293bc69
Create Date: 2025-04-04 09:21:58.070532

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "7a73514ca2d6"
down_revision = "06b7c293bc69"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_index("ix_events__event_related_occurred", table_name="events")
    op.drop_index("ix_events__related_resource_ids", table_name="events")
    op.create_index(
        "ix_events__related_gin",
        "events",
        ["related"],
        unique=False,
        postgresql_using="gin",
    )
    op.create_index(
        "ix_events__event_occurred",
        "events",
        ["event", "occurred"],
        unique=False,
    )
    op.create_index(
        "ix_events__related_resource_ids_gin",
        "events",
        ["related_resource_ids"],
        unique=False,
        postgresql_using="gin",
    )


def downgrade():
    op.drop_index(
        "ix_events__related_gin",
        table_name="events",
    )
    op.drop_index(
        "ix_events__event_occurred",
        table_name="events",
    )
    op.drop_index(
        "ix_events__related_resource_ids_gin",
        table_name="events",
    )
    op.create_index(
        "ix_events__event_related_occurred",
        "events",
        ["event", "related", "occurred"],
        unique=False,
    )
    op.create_index(
        "ix_events__related_resource_ids",
        "events",
        ["related_resource_ids"],
        unique=False,
    )



from . import (
    states,
    statuses,
    schedules,
    core,
    sorting,
    filters,
    responses,
    actions,
    internal,
    ui,
)  # noqa

__all__ = [
    "actions",
    "core",
    "filters",
    "responses",
    "schedules",
    "sorting",
    "states",
    "statuses",
    "internal",
    "ui",
]



from __future__ import annotations

import inspect
import logging
from typing import Any, Optional

import typer

from prefect.cli._types import PrefectTyper
from prefect.cli._utilities import exit_with_error
from prefect.cli.root import app
from prefect.logging import get_logger
from prefect.task_worker import serve as task_serve
from prefect.tasks import Task
from prefect.utilities.importtools import import_object, load_module

task_app: PrefectTyper = PrefectTyper(name="task", help="Work with task scheduling.")
app.add_typer(task_app, aliases=["task"])

logger: logging.Logger = get_logger("prefect.cli.task")


def _import_tasks_from_module(module: str) -> list[Task[Any, Any]]:
    mod = load_module(module)
    return [
        obj
        for _, obj in inspect.getmembers(mod)
        if isinstance(obj, Task) and not inspect.ismodule(obj)
    ]


@task_app.command()
async def serve(
    entrypoints: Optional[list[str]] = typer.Argument(
        None,
        help="The paths to one or more tasks, in the form of `./path/to/file.py:task_func_name`.",
    ),
    module: Optional[list[str]] = typer.Option(
        None,
        "--module",
        "-m",
        help="The module(s) to import the tasks from.",
    ),
    limit: int = typer.Option(
        10,
        help="The maximum number of tasks that can be run concurrently. Defaults to 10.",
    ),
):
    """
    Serve the provided tasks so that their runs may be submitted to and
    executed in the engine.

    Args:
        entrypoints: List of strings representing the paths to one or more
            tasks. Each path should be in the format
            `./path/to/file.py:task_func_name`.
        module: The module(s) to import the task definitions from.
        limit: The maximum number of tasks that can be run concurrently. Defaults to 10.
    """
    if (entrypoints and any(entrypoints)) and (module and any(module)):
        exit_with_error(
            "You may provide entrypoints or modules, but not both at the same time."
        )

    if not entrypoints and not module:
        exit_with_error(
            "You must provide either `path/to/file.py:task_func` or `--module module_name`."
        )

    tasks: list[Task[Any, Any]] = []

    if entrypoints:
        for entrypoint in entrypoints:
            if ".py:" not in entrypoint:
                exit_with_error(
                    (
                        f"Error: Invalid entrypoint format {entrypoint!r}. It "
                        "must be of the form `./path/to/file.py:task_func_name`."
                    )
                )

            try:
                tasks.append(import_object(entrypoint))
            except Exception:
                mod, task_name = entrypoint.split(":")
                exit_with_error(
                    f"Error: {mod!r} has no function {task_name!r}.", style="red"
                )

    elif module:
        for mod in module:
            try:
                module_tasks = _import_tasks_from_module(mod)
            except Exception as e:
                exit_with_error(
                    f"Module '{mod}' could not be imported. Please check the module name and try again.\n\n{e.__class__.__name__}: {e}"
                )
            plural = "s" if len(module_tasks) != 1 else ""
            logger.debug(f"Found {len(module_tasks)} task{plural} in {mod!r}.")
            tasks.extend(module_tasks)

    if not tasks:
        sources = (
            f"entrypoints: {entrypoints!r}" if entrypoints else f"modules: {module!r}"
        )
        exit_with_error(f"No tasks found to serve in {sources}.")

    await task_serve(*tasks, limit=limit)



import datetime
import time
from datetime import timedelta
from logging import Logger
from pathlib import Path
from typing import Optional

import anyio
import pydantic_core
from typing_extensions import TypedDict

from prefect.logging.loggers import get_logger
from prefect.types._datetime import now, parse_datetime

from .protocol import LockManager

logger: Logger = get_logger(__name__)


class _LockInfo(TypedDict):
    """
    A dictionary containing information about a lock.

    Attributes:
        holder: The holder of the lock.
        expiration: Datetime when the lock expires.
        path: Path to the lock file.
    """

    holder: str
    expiration: Optional[datetime.datetime]
    path: Path


class FileSystemLockManager(LockManager):
    """
    A lock manager that implements locking using local files.

    Attributes:
        lock_files_directory: the directory where lock files are stored
    """

    def __init__(self, lock_files_directory: Path) -> None:
        self.lock_files_directory: Path = lock_files_directory.expanduser().resolve()
        self._locks: dict[str, _LockInfo] = {}

    def _ensure_lock_files_directory_exists(self) -> None:
        self.lock_files_directory.mkdir(parents=True, exist_ok=True)

    def _lock_path_for_key(self, key: str) -> Path:
        if (lock_info := self._locks.get(key)) is not None:
            return lock_info["path"]
        return self.lock_files_directory.joinpath(key).with_suffix(".lock")

    def _get_lock_info(self, key: str, use_cache: bool = True) -> Optional[_LockInfo]:
        if use_cache:
            if (lock_info := self._locks.get(key)) is not None:
                return lock_info

        lock_path = self._lock_path_for_key(key)

        try:
            with open(lock_path, "rb") as lock_file:
                lock_info = pydantic_core.from_json(lock_file.read())
                lock_info["path"] = lock_path
                expiration = lock_info.get("expiration")
                lock_info["expiration"] = (
                    parse_datetime(expiration) if expiration is not None else None
                )
            self._locks[key] = lock_info
            return lock_info
        except FileNotFoundError:
            return None

    async def _aget_lock_info(
        self, key: str, use_cache: bool = True
    ) -> Optional[_LockInfo]:
        if use_cache:
            if (lock_info := self._locks.get(key)) is not None:
                return lock_info

        lock_path = self._lock_path_for_key(key)

        try:
            lock_info_bytes = await anyio.Path(lock_path).read_bytes()
            lock_info = pydantic_core.from_json(lock_info_bytes)
            lock_info["path"] = lock_path
            expiration = lock_info.get("expiration")
            lock_info["expiration"] = (
                parse_datetime(expiration) if expiration is not None else None
            )
            self._locks[key] = lock_info
            return lock_info
        except FileNotFoundError:
            return None

    def acquire_lock(
        self,
        key: str,
        holder: str,
        acquire_timeout: Optional[float] = None,
        hold_timeout: Optional[float] = None,
    ) -> bool:
        self._ensure_lock_files_directory_exists()
        lock_path = self._lock_path_for_key(key)

        if self.is_locked(key) and not self.is_lock_holder(key, holder):
            lock_free = self.wait_for_lock(key, acquire_timeout)
            if not lock_free:
                return False

        try:
            Path(lock_path).touch(exist_ok=False)
        except FileExistsError:
            if not self.is_lock_holder(key, holder):
                logger.debug(
                    f"Another actor acquired the lock for record with key {key}. Trying again."
                )
                return self.acquire_lock(key, holder, acquire_timeout, hold_timeout)
        expiration = (
            now("UTC") + timedelta(seconds=hold_timeout)
            if hold_timeout is not None
            else None
        )

        with open(Path(lock_path), "wb") as lock_file:
            lock_file.write(
                pydantic_core.to_json(
                    {
                        "holder": holder,
                        "expiration": str(expiration)
                        if expiration is not None
                        else None,
                    },
                )
            )

        self._locks[key] = {
            "holder": holder,
            "expiration": expiration,
            "path": lock_path,
        }

        return True

    async def aacquire_lock(
        self,
        key: str,
        holder: str,
        acquire_timeout: Optional[float] = None,
        hold_timeout: Optional[float] = None,
    ) -> bool:
        await anyio.Path(self.lock_files_directory).mkdir(parents=True, exist_ok=True)
        lock_path = self._lock_path_for_key(key)

        if self.is_locked(key) and not self.is_lock_holder(key, holder):
            lock_free = await self.await_for_lock(key, acquire_timeout)
            if not lock_free:
                return False

        try:
            await anyio.Path(lock_path).touch(exist_ok=False)
        except FileExistsError:
            if not self.is_lock_holder(key, holder):
                logger.debug(
                    f"Another actor acquired the lock for record with key {key}. Trying again."
                )
                return self.acquire_lock(key, holder, acquire_timeout, hold_timeout)
        expiration = (
            now("UTC") + timedelta(seconds=hold_timeout)
            if hold_timeout is not None
            else None
        )

        async with await anyio.Path(lock_path).open("wb") as lock_file:
            await lock_file.write(
                pydantic_core.to_json(
                    {
                        "holder": holder,
                        "expiration": str(expiration)
                        if expiration is not None
                        else None,
                    },
                )
            )

        self._locks[key] = {
            "holder": holder,
            "expiration": expiration,
            "path": lock_path,
        }

        return True

    def release_lock(self, key: str, holder: str) -> None:
        lock_path = self._lock_path_for_key(key)
        if not self.is_locked(key):
            ValueError(f"No lock for transaction with key {key}")
        if self.is_lock_holder(key, holder):
            Path(lock_path).unlink(missing_ok=True)
            self._locks.pop(key, None)
        else:
            raise ValueError(f"No lock held by {holder} for transaction with key {key}")

    def is_locked(self, key: str, use_cache: bool = False) -> bool:
        if (lock_info := self._get_lock_info(key, use_cache=use_cache)) is None:
            return False

        if (expiration := lock_info.get("expiration")) is None:
            return True

        expired = expiration < now("UTC")
        if expired:
            Path(lock_info["path"]).unlink()
            self._locks.pop(key, None)
            return False
        else:
            return True

    def is_lock_holder(self, key: str, holder: str) -> bool:
        if not self.is_locked(key):
            return False

        if not self.is_locked(key):
            return False
        if (lock_info := self._get_lock_info(key)) is None:
            return False
        return lock_info["holder"] == holder

    def wait_for_lock(self, key: str, timeout: Optional[float] = None) -> bool:
        seconds_waited = 0
        while self.is_locked(key, use_cache=False):
            if timeout and seconds_waited >= timeout:
                return False
            seconds_waited += 0.1
            time.sleep(0.1)
        return True

    async def await_for_lock(self, key: str, timeout: Optional[float] = None) -> bool:
        seconds_waited = 0
        while self.is_locked(key, use_cache=False):
            if timeout and seconds_waited >= timeout:
                return False
            seconds_waited += 0.1
            await anyio.sleep(0.1)
        return True



"""Add block_type_name to block document

Revision ID: cef24af2ec34
Revises: f3165ae0a213
Create Date: 2023-10-30 07:50:26.414043

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "cef24af2ec34"
down_revision = "f3165ae0a213"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("block_document", schema=None) as batch_op:
        batch_op.add_column(sa.Column("block_type_name", sa.String(), nullable=True))


def downgrade():
    with op.batch_alter_table("block_document", schema=None) as batch_op:
        batch_op.drop_column("block_type_name")



"""
Command line interface for working with work queues.
"""

import datetime
import warnings
from textwrap import dedent
from typing import Optional, Union
from uuid import UUID

import typer
from rich.pretty import Pretty
from rich.table import Table

from prefect.cli._types import PrefectTyper
from prefect.cli._utilities import exit_with_error, exit_with_success
from prefect.cli.root import app, is_interactive
from prefect.client.orchestration import get_client
from prefect.client.schemas.filters import WorkPoolFilter, WorkPoolFilterId
from prefect.client.schemas.objects import (
    DEFAULT_AGENT_WORK_POOL_NAME,
    FlowRun,
    WorkQueue,
)
from prefect.exceptions import ObjectAlreadyExists, ObjectNotFound
from prefect.types._datetime import now as now_fn

work_app: PrefectTyper = PrefectTyper(name="work-queue", help="Manage work queues.")
app.add_typer(work_app, aliases=["work-queues"])


async def _get_work_queue_id_from_name_or_id(
    name_or_id: Union[UUID, str], work_pool_name: Optional[str] = None
) -> UUID:
    """
    For backwards-compatibility, the main argument of the work queue CLI can be
    either a name (preferred) or an ID (legacy behavior).
    """

    if not name_or_id:
        # hint that we prefer names
        exit_with_error("Provide a work queue name.")

    # try parsing as ID
    try:
        return UUID(name_or_id)

    # parse as string
    except (AttributeError, ValueError):
        async with get_client() as client:
            try:
                work_queue = await client.read_work_queue_by_name(
                    name=name_or_id,
                    work_pool_name=work_pool_name,
                )
                return work_queue.id
            except ObjectNotFound:
                if not work_pool_name:
                    exit_with_error(f"No work queue named {name_or_id!r} found.")

                exit_with_error(
                    f"No work queue named {name_or_id!r} found in work pool"
                    f" {work_pool_name!r}."
                )


@work_app.command()
async def create(
    name: str = typer.Argument(..., help="The unique name to assign this work queue"),
    limit: int = typer.Option(
        None, "-l", "--limit", help="The concurrency limit to set on the queue."
    ),
    pool: Optional[str] = typer.Option(
        None,
        "-p",
        "--pool",
        help="The name of the work pool to create the work queue in.",
    ),
    priority: Optional[int] = typer.Option(
        None,
        "-q",
        "--priority",
        help="The associated priority for the created work queue",
    ),
):
    """
    Create a work queue.
    """

    async with get_client() as client:
        try:
            result = await client.create_work_queue(
                name=name, work_pool_name=pool, priority=priority
            )
            if limit is not None:
                await client.update_work_queue(
                    id=result.id,
                    concurrency_limit=limit,
                )
        except ObjectAlreadyExists:
            exit_with_error(f"Work queue with name: {name!r} already exists.")
        except ObjectNotFound:
            exit_with_error(f"Work pool with name: {pool!r} not found.")

    if not pool:
        # specify the default work pool name after work queue creation to allow the server
        # to handle a bunch of logic associated with agents without work pools
        pool = DEFAULT_AGENT_WORK_POOL_NAME
    output_msg = dedent(
        f"""
        Created work queue with properties:
            name - {name!r}
            work pool - {pool!r}
            id - {result.id}
            concurrency limit - {limit}
        Start a worker to pick up flow runs from the work queue:
            prefect worker start -q '{result.name} -p {pool}'

        Inspect the work queue:
            prefect work-queue inspect '{result.name}'
        """
    )
    exit_with_success(output_msg)


@work_app.command()
async def set_concurrency_limit(
    name: str = typer.Argument(..., help="The name or ID of the work queue"),
    limit: int = typer.Argument(..., help="The concurrency limit to set on the queue."),
    pool: Optional[str] = typer.Option(
        None,
        "-p",
        "--pool",
        help="The name of the work pool that the work queue belongs to.",
    ),
):
    """
    Set a concurrency limit on a work queue.
    """
    queue_id = await _get_work_queue_id_from_name_or_id(
        name_or_id=name,
        work_pool_name=pool,
    )

    async with get_client() as client:
        try:
            await client.update_work_queue(
                id=queue_id,
                concurrency_limit=limit,
            )
        except ObjectNotFound:
            if pool:
                error_message = (
                    f"No work queue named {name!r} found in work pool {pool!r}."
                )
            else:
                error_message = f"No work queue named {name!r} found."
            exit_with_error(error_message)

    if pool:
        success_message = (
            f"Concurrency limit of {limit} set on work queue {name!r} in work pool"
            f" {pool!r}"
        )
    else:
        success_message = f"Concurrency limit of {limit} set on work queue {name!r}"
    exit_with_success(success_message)


@work_app.command()
async def clear_concurrency_limit(
    name: str = typer.Argument(..., help="The name or ID of the work queue to clear"),
    pool: Optional[str] = typer.Option(
        None,
        "-p",
        "--pool",
        help="The name of the work pool that the work queue belongs to.",
    ),
):
    """
    Clear any concurrency limits from a work queue.
    """
    queue_id = await _get_work_queue_id_from_name_or_id(
        name_or_id=name,
        work_pool_name=pool,
    )
    async with get_client() as client:
        try:
            await client.update_work_queue(
                id=queue_id,
                concurrency_limit=None,
            )
        except ObjectNotFound:
            if pool:
                error_message = f"No work queue found: {name!r} in work pool {pool!r}"
            else:
                error_message = f"No work queue found: {name!r}"
            exit_with_error(error_message)

    if pool:
        success_message = (
            f"Concurrency limits removed on work queue {name!r} in work pool {pool!r}"
        )
    else:
        success_message = f"Concurrency limits removed on work queue {name!r}"
    exit_with_success(success_message)


@work_app.command()
async def pause(
    name: str = typer.Argument(..., help="The name or ID of the work queue to pause"),
    pool: Optional[str] = typer.Option(
        None,
        "-p",
        "--pool",
        help="The name of the work pool that the work queue belongs to.",
    ),
):
    """
    Pause a work queue.
    """

    if not pool and not typer.confirm(
        f"You have not specified a work pool. Are you sure you want to pause {name} work queue in `{DEFAULT_AGENT_WORK_POOL_NAME}`?"
    ):
        exit_with_error("Work queue pause aborted!")

    queue_id = await _get_work_queue_id_from_name_or_id(
        name_or_id=name,
        work_pool_name=pool,
    )

    async with get_client() as client:
        try:
            await client.update_work_queue(
                id=queue_id,
                is_paused=True,
            )
        except ObjectNotFound:
            if pool:
                error_message = f"No work queue found: {name!r} in work pool {pool!r}"
            else:
                error_message = f"No work queue found: {name!r}"
            exit_with_error(error_message)

    if pool:
        success_message = f"Work queue {name!r} in work pool {pool!r} paused"
    else:
        success_message = f"Work queue {name!r} paused"
    exit_with_success(success_message)


@work_app.command()
async def resume(
    name: str = typer.Argument(..., help="The name or ID of the work queue to resume"),
    pool: Optional[str] = typer.Option(
        None,
        "-p",
        "--pool",
        help="The name of the work pool that the work queue belongs to.",
    ),
):
    """
    Resume a paused work queue.
    """
    queue_id = await _get_work_queue_id_from_name_or_id(
        name_or_id=name,
        work_pool_name=pool,
    )

    async with get_client() as client:
        try:
            await client.update_work_queue(
                id=queue_id,
                is_paused=False,
            )
        except ObjectNotFound:
            if pool:
                error_message = f"No work queue found: {name!r} in work pool {pool!r}"
            else:
                error_message = f"No work queue found: {name!r}"
            exit_with_error(error_message)

    if pool:
        success_message = f"Work queue {name!r} in work pool {pool!r} resumed"
    else:
        success_message = f"Work queue {name!r} resumed"
    exit_with_success(success_message)


@work_app.command()
async def inspect(
    name: str = typer.Argument(
        None, help="The name or ID of the work queue to inspect"
    ),
    pool: Optional[str] = typer.Option(
        None,
        "-p",
        "--pool",
        help="The name of the work pool that the work queue belongs to.",
    ),
):
    """
    Inspect a work queue by ID.
    """
    queue_id = await _get_work_queue_id_from_name_or_id(
        name_or_id=name,
        work_pool_name=pool,
    )
    async with get_client() as client:
        try:
            result = await client.read_work_queue(id=queue_id)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=DeprecationWarning)

                app.console.print(Pretty(result))
        except ObjectNotFound:
            if pool:
                error_message = f"No work queue found: {name!r} in work pool {pool!r}"
            else:
                error_message = f"No work queue found: {name!r}"
            exit_with_error(error_message)

        try:
            status = await client.read_work_queue_status(id=queue_id)
            app.console.print(Pretty(status))
        except ObjectNotFound:
            pass


@work_app.command()
async def ls(
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Display more information."
    ),
    work_queue_prefix: str = typer.Option(
        None,
        "--match",
        "-m",
        help=(
            "Will match work queues with names that start with the specified prefix"
            " string"
        ),
    ),
    pool: Optional[str] = typer.Option(
        None,
        "-p",
        "--pool",
        help="The name of the work pool containing the work queues to list.",
    ),
):
    """
    View all work queues.
    """
    if not pool:
        table = Table(
            title="Work Queues",
            caption="(**) denotes a paused queue",
            caption_style="red",
        )
        table.add_column("Name", style="green", no_wrap=True)
        table.add_column("Pool", style="magenta", no_wrap=True)
        table.add_column("ID", justify="right", style="cyan", no_wrap=True)
        table.add_column("Concurrency Limit", style="blue", no_wrap=True)
        if verbose:
            table.add_column("Filter (Deprecated)", style="magenta", no_wrap=True)

        async with get_client() as client:
            if work_queue_prefix is not None:
                queues = await client.match_work_queues([work_queue_prefix])
            else:
                queues = await client.read_work_queues()

            pool_ids = [q.work_pool_id for q in queues]
            wp_filter = WorkPoolFilter(id=WorkPoolFilterId(any_=pool_ids))
            pools = await client.read_work_pools(work_pool_filter=wp_filter)
            pool_id_name_map = {p.id: p.name for p in pools}

            def sort_by_created_key(q: WorkQueue):
                assert q.created is not None, "created is not None"
                return now_fn("UTC") - q.created

            for queue in sorted(queues, key=sort_by_created_key):
                row = [
                    f"{queue.name} [red](**)" if queue.is_paused else queue.name,
                    pool_id_name_map[queue.work_pool_id],
                    str(queue.id),
                    (
                        f"[red]{queue.concurrency_limit}"
                        if queue.concurrency_limit is not None
                        else "[blue]None"
                    ),
                ]
                if verbose and queue.filter is not None:
                    row.append(queue.filter.model_dump_json())
                table.add_row(*row)

    else:
        table = Table(
            title=f"Work Queues in Work Pool {pool!r}",
            caption="(**) denotes a paused queue",
            caption_style="red",
        )
        table.add_column("Name", style="green", no_wrap=True)
        table.add_column("Priority", style="magenta", no_wrap=True)
        table.add_column("Concurrency Limit", style="blue", no_wrap=True)
        if verbose:
            table.add_column("Description", style="cyan", no_wrap=False)

        async with get_client() as client:
            try:
                queues = await client.read_work_queues(work_pool_name=pool)
            except ObjectNotFound:
                exit_with_error(f"No work pool found: {pool!r}")

            def sort_by_created_key(q: WorkQueue):
                assert q.created is not None, "created is not None"
                return now_fn("UTC") - q.created

            for queue in sorted(queues, key=sort_by_created_key):
                row = [
                    f"{queue.name} [red](**)" if queue.is_paused else queue.name,
                    f"{queue.priority}",
                    (
                        f"[red]{queue.concurrency_limit}"
                        if queue.concurrency_limit is not None
                        else "[blue]None"
                    ),
                ]
                if verbose:
                    row.append(queue.description)
                table.add_row(*row)

    app.console.print(table)


@work_app.command()
async def preview(
    name: str = typer.Argument(
        None, help="The name or ID of the work queue to preview"
    ),
    hours: int = typer.Option(
        None,
        "-h",
        "--hours",
        help="The number of hours to look ahead; defaults to 1 hour",
    ),
    pool: Optional[str] = typer.Option(
        None,
        "-p",
        "--pool",
        help="The name of the work pool that the work queue belongs to.",
    ),
):
    """
    Preview a work queue.
    """
    if pool:
        title = f"Preview of Work Queue {name!r} in Work Pool {pool!r}"
    else:
        title = f"Preview of Work Queue {name!r}"

    table = Table(title=title, caption="(**) denotes a late run", caption_style="red")
    table.add_column(
        "Scheduled Start Time", justify="left", style="yellow", no_wrap=True
    )
    table.add_column("Run ID", justify="left", style="cyan", no_wrap=True)
    table.add_column("Name", style="green", no_wrap=True)
    table.add_column("Deployment ID", style="blue", no_wrap=True)

    window = now_fn("UTC") + datetime.timedelta(hours=hours or 1)

    queue_id = await _get_work_queue_id_from_name_or_id(
        name_or_id=name, work_pool_name=pool
    )
    async with get_client() as client:
        if pool:
            try:
                responses = await client.get_scheduled_flow_runs_for_work_pool(
                    work_pool_name=pool,
                    work_queue_names=[name],
                )
                runs = [response.flow_run for response in responses]
            except ObjectNotFound:
                exit_with_error(f"No work queue found: {name!r} in work pool {pool!r}")
        else:
            try:
                runs = await client.get_runs_in_work_queue(
                    queue_id,
                    limit=10,
                    scheduled_before=window,
                )
            except ObjectNotFound:
                exit_with_error(f"No work queue found: {name!r}")
    now = now_fn("UTC")

    def sort_by_created_key(r: FlowRun):
        assert r.created is not None, "created is not None"
        return now - r.created

    for run in sorted(runs, key=sort_by_created_key):
        table.add_row(
            (
                f"{run.expected_start_time} [red](**)"
                if run.expected_start_time < now
                else f"{run.expected_start_time}"
            ),
            str(run.id),
            run.name,
            str(run.deployment_id),
        )

    if runs:
        app.console.print(table)
    else:
        app.console.print(
            (
                "No runs found - try increasing how far into the future you preview"
                " with the --hours flag"
            ),
            style="yellow",
        )


@work_app.command()
async def delete(
    name: str = typer.Argument(..., help="The name or ID of the work queue to delete"),
    pool: Optional[str] = typer.Option(
        None,
        "-p",
        "--pool",
        help="The name of the work pool containing the work queue to delete.",
    ),
):
    """
    Delete a work queue by ID.
    """

    queue_id = await _get_work_queue_id_from_name_or_id(
        name_or_id=name,
        work_pool_name=pool,
    )
    async with get_client() as client:
        try:
            if is_interactive() and not typer.confirm(
                (f"Are you sure you want to delete work queue with name {name!r}?"),
                default=False,
            ):
                exit_with_error("Deletion aborted.")
            await client.delete_work_queue_by_id(id=queue_id)
        except ObjectNotFound:
            if pool:
                error_message = f"No work queue found: {name!r} in work pool {pool!r}"
            else:
                error_message = f"No work queue found: {name!r}"
            exit_with_error(error_message)
    if pool:
        success_message = (
            f"Successfully deleted work queue {name!r} in work pool {pool!r}"
        )
    else:
        success_message = f"Successfully deleted work queue {name!r}"
    exit_with_success(success_message)


@work_app.command("read-runs")
async def read_wq_runs(
    name: str = typer.Argument(..., help="The name or ID of the work queue to poll"),
    pool: Optional[str] = typer.Option(
        None,
        "-p",
        "--pool",
        help="The name of the work pool containing the work queue to poll.",
    ),
):
    """
    Get runs in a work queue. Note that this will trigger an artificial poll of
    the work queue.
    """

    queue_id = await _get_work_queue_id_from_name_or_id(
        name_or_id=name,
        work_pool_name=pool,
    )
    async with get_client() as client:
        try:
            runs = await client.get_runs_in_work_queue(id=queue_id)
        except ObjectNotFound:
            if pool:
                error_message = f"No work queue found: {name!r} in work pool {pool!r}"
            else:
                error_message = f"No work queue found: {name!r}"
            exit_with_error(error_message)
    success_message = (
        f"Read {len(runs)} runs for work queue {name!r} in work pool {pool}: {runs}"
    )
    exit_with_success(success_message)



"""Add job_variables column to flow-runs

Revision ID: 342220764f0b
Revises: 265eb1a2da4c
Create Date: 2024-03-05 11:52:58.330563

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Text

import prefect

# revision identifiers, used by Alembic.
revision = "342220764f0b"
down_revision = "265eb1a2da4c"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "flow_run",
        sa.Column(
            "job_variables",
            prefect.server.utilities.database.JSON(astext_type=Text()),
            server_default="{}",
            nullable=True,
        ),
    )


def downgrade():
    op.drop_column("flow_run", "job_variables")



"""
Command line interface for working with concurrency limits.
"""

import textwrap

import typer
from rich.console import Group
from rich.panel import Panel
from rich.pretty import Pretty
from rich.table import Table

from prefect.cli._types import PrefectTyper
from prefect.cli._utilities import exit_with_error, exit_with_success
from prefect.cli.root import app, is_interactive
from prefect.client.orchestration import get_client
from prefect.exceptions import ObjectNotFound
from prefect.types._datetime import human_friendly_diff

concurrency_limit_app: PrefectTyper = PrefectTyper(
    name="concurrency-limit",
    help="Manage task-level concurrency limits.",
)
app.add_typer(concurrency_limit_app, aliases=["concurrency-limits"])


@concurrency_limit_app.command()
async def create(tag: str, concurrency_limit: int):
    """
    Create a concurrency limit against a tag.

    This limit controls how many task runs with that tag may simultaneously be in a
    Running state.
    """

    async with get_client() as client:
        await client.create_concurrency_limit(
            tag=tag, concurrency_limit=concurrency_limit
        )
        await client.read_concurrency_limit_by_tag(tag)

    app.console.print(
        textwrap.dedent(
            f"""
            Created concurrency limit with properties:
                tag - {tag!r}
                concurrency_limit - {concurrency_limit}

            Delete the concurrency limit:
                prefect concurrency-limit delete {tag!r}
            
            Inspect the concurrency limit:
                prefect concurrency-limit inspect {tag!r}
        """
        )
    )


@concurrency_limit_app.command()
async def inspect(tag: str):
    """
    View details about a concurrency limit. `active_slots` shows a list of TaskRun IDs
    which are currently using a concurrency slot.
    """

    async with get_client() as client:
        try:
            result = await client.read_concurrency_limit_by_tag(tag=tag)
        except ObjectNotFound:
            exit_with_error(f"No concurrency limit found for the tag: {tag}")

    trid_table = Table()
    trid_table.add_column("Active Task Run IDs", style="cyan", no_wrap=True)

    cl_table = Table(title=f"Concurrency Limit ID: [red]{str(result.id)}")
    cl_table.add_column("Tag", style="green", no_wrap=True)
    cl_table.add_column("Concurrency Limit", style="blue", no_wrap=True)
    cl_table.add_column("Created", style="magenta", no_wrap=True)
    cl_table.add_column("Updated", style="magenta", no_wrap=True)

    for trid in sorted(result.active_slots):
        trid_table.add_row(str(trid))

    cl_table.add_row(
        str(result.tag),
        str(result.concurrency_limit),
        Pretty(human_friendly_diff(result.created) if result.created else ""),
        Pretty(human_friendly_diff(result.updated) if result.updated else ""),
    )

    group = Group(
        cl_table,
        trid_table,
    )
    app.console.print(Panel(group, expand=False))


@concurrency_limit_app.command()
async def ls(limit: int = 15, offset: int = 0):
    """
    View all concurrency limits.
    """
    table = Table(
        title="Concurrency Limits",
        caption="inspect a concurrency limit to show active task run IDs",
    )
    table.add_column("Tag", style="green", no_wrap=True)
    table.add_column("ID", justify="right", style="cyan", no_wrap=True)
    table.add_column("Concurrency Limit", style="blue", no_wrap=True)
    table.add_column("Active Task Runs", style="magenta", no_wrap=True)

    async with get_client() as client:
        concurrency_limits = await client.read_concurrency_limits(
            limit=limit, offset=offset
        )

    for cl in sorted(
        concurrency_limits, key=lambda c: c.updated or c.created or "", reverse=True
    ):
        table.add_row(
            str(cl.tag),
            str(cl.id),
            str(cl.concurrency_limit),
            str(len(cl.active_slots)),
        )

    app.console.print(table)


@concurrency_limit_app.command()
async def reset(tag: str):
    """
    Resets the concurrency limit slots set on the specified tag.
    """

    async with get_client() as client:
        try:
            await client.reset_concurrency_limit_by_tag(tag=tag)
        except ObjectNotFound:
            exit_with_error(f"No concurrency limit found for the tag: {tag}")

    exit_with_success(f"Reset concurrency limit set on the tag: {tag}")


@concurrency_limit_app.command()
async def delete(tag: str):
    """
    Delete the concurrency limit set on the specified tag.
    """

    async with get_client() as client:
        try:
            if is_interactive() and not typer.confirm(
                (
                    f"Are you sure you want to delete concurrency limit with tag {tag!r}?"
                ),
                default=False,
            ):
                exit_with_error("Deletion aborted.")
            await client.delete_concurrency_limit_by_tag(tag=tag)
        except ObjectNotFound:
            exit_with_error(f"No concurrency limit found for the tag: {tag}")

    exit_with_success(f"Deleted concurrency limit set on the tag: {tag}")



"""add scope+leader idx to AutomationEventFollower

Revision ID: 1c9bb7f78263
Revises: 4160a4841eed
Create Date: 2025-05-04 19:32:49.195966

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "1c9bb7f78263"
down_revision = "4160a4841eed"
branch_labels = None
depends_on = None


def upgrade():
    op.create_index(
        index_name="ix_ae_follower_scope_leader",
        table_name="automation_event_follower",
        columns=["scope", "leader_event_id"],
    )


def downgrade():
    op.drop_index(index_name="ix_ae_follower_scope_leader")



"""Sync ORM models and migrations

Revision ID: a49711513ad4
Revises: 5952a5498b51
Create Date: 2024-12-04 14:49:24.099491

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import sqlite

# revision identifiers, used by Alembic.
revision = "a49711513ad4"
down_revision = "5952a5498b51"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("artifact_collection", schema=None) as batch_op:
        # table added in 027c123512befd2bd00a0ef28bd44215e77bece6 but index was
        # never created in a migration.
        batch_op.create_index(
            batch_op.f("ix_artifact_collection__updated"), ["updated"], unique=False
        )
        # index created on the wrong table in ca9f93463a4c38fce8be972d91e808b5935e5d9c
        batch_op.drop_index("ix_artifact__key_created_desc")

    with op.batch_alter_table("artifact", schema=None) as batch_op:
        # index created on the wrong table in ca9f93463a4c38fce8be972d91e808b5935e5d9c
        batch_op.create_index(
            "ix_artifact__key_created_desc",
            ["key", sa.text("created DESC")],
            unique=False,
            postgresql_include=["id", "updated", "type", "task_run_id", "flow_run_id"],
        )

    with op.batch_alter_table("block_document", schema=None) as batch_op:
        # Renamed index to remain consistent with PostgreSQL
        batch_op.drop_index("ix_block_document__block_type_name_name")
        batch_op.create_index(
            "ix_block_document__block_type_name__name",
            ["block_type_name", "name"],
            unique=False,
        )

    with op.batch_alter_table("deployment", schema=None) as batch_op:
        # columns removed in c53b00bfa1f6850ab43e168c92c627350c090647
        batch_op.drop_column("schedule")
        batch_op.drop_column("is_schedule_active")

        # column removed in 5784c637e7e11a8e88e2b3146e54e9b6c97d50ef
        batch_op.drop_column("flow_data")

        # column removed in eaa7a5063c73718dff56ce4aeb66e53fcafe60e5
        batch_op.drop_column("manifest_path")

    with op.batch_alter_table("deployment_schedule", schema=None) as batch_op:
        # columns removed from orm models in 0b62de684447c6955e04c722c276edac4002fd40
        batch_op.drop_column("catchup")
        batch_op.drop_column("max_active_runs")

    with op.batch_alter_table("flow_run", schema=None) as batch_op:
        # Column is no longer a FK since d10c7471a69403bcf88f401091497a2dc8963885
        batch_op.drop_index("ix_flow_run__deployment_id")
        # Index accidentally dropped in 519a2ed6e31e2b60136e1a1a163a9cd0a8d3d5c4
        batch_op.create_index(
            "ix_flow_run__scheduler_deployment_id_auto_scheduled_next_schedu",
            ["deployment_id", "auto_scheduled", "next_scheduled_start_time"],
            unique=False,
            postgresql_where=sa.text("state_type = 'SCHEDULED'::state_type"),
            sqlite_where=sa.text("state_type = 'SCHEDULED'"),
        )


def downgrade():
    with op.batch_alter_table("flow_run", schema=None) as batch_op:
        batch_op.drop_index(
            "ix_flow_run__scheduler_deployment_id_auto_scheduled_next_schedu",
            postgresql_where=sa.text("state_type = 'SCHEDULED'::state_type"),
            sqlite_where=sa.text("state_type = 'SCHEDULED'"),
        )
        batch_op.create_index(
            "ix_flow_run__deployment_id", ["deployment_id"], unique=False
        )

    with op.batch_alter_table("deployment_schedule", schema=None) as batch_op:
        batch_op.add_column(sa.Column("max_active_runs", sa.INTEGER(), nullable=True))
        batch_op.add_column(
            sa.Column(
                "catchup", sa.BOOLEAN(), server_default=sa.text("'0'"), nullable=False
            )
        )

    with op.batch_alter_table("deployment", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "is_schedule_active",
                sa.BOOLEAN(),
                server_default=sa.text("'1'"),
                nullable=False,
            )
        )
        batch_op.add_column(sa.Column("flow_data", sqlite.JSON(), nullable=True))
        batch_op.add_column(sa.Column("schedule", sqlite.JSON(), nullable=True))
        batch_op.add_column(sa.Column("manifest_path", sa.VARCHAR(), nullable=True))

    with op.batch_alter_table("block_document", schema=None) as batch_op:
        batch_op.drop_index("ix_block_document__block_type_name__name")
        batch_op.create_index(
            "ix_block_document__block_type_name_name",
            ["block_type_name", "name"],
            unique=False,
        )

    with op.batch_alter_table("artifact", schema=None) as batch_op:
        batch_op.drop_index(
            "ix_artifact__key_created_desc",
            postgresql_include=["id", "updated", "type", "task_run_id", "flow_run_id"],
        )

    with op.batch_alter_table("artifact_collection", schema=None) as batch_op:
        batch_op.create_index(
            "ix_artifact__key_created_desc", ["key", "created"], unique=False
        )
        batch_op.drop_index(batch_op.f("ix_artifact_collection__updated"))



import warnings
from typing import Any, ClassVar, Optional
from urllib.parse import quote_plus

from pydantic import (
    AliasChoices,
    AliasPath,
    Field,
    SecretStr,
    model_validator,
)
from pydantic_settings import SettingsConfigDict
from typing_extensions import Literal, Self

from prefect.settings.base import PrefectBaseSettings, build_settings_config


class SQLAlchemyTLSSettings(PrefectBaseSettings):
    """
    Settings for controlling SQLAlchemy mTLS context when
    using a PostgreSQL database.
    """

    model_config: ClassVar[SettingsConfigDict] = build_settings_config(
        ("server", "database", "sqlalchemy", "connect_args", "tls")
    )

    enabled: bool = Field(
        default=False,
        description="Controls whether connected to mTLS enabled PostgreSQL when using a PostgreSQL database with the Prefect backend.",
    )

    ca_file: Optional[str] = Field(
        default=None,
        description="This configuration settings option specifies the path to PostgreSQL client certificate authority file.",
    )

    cert_file: Optional[str] = Field(
        default=None,
        description="This configuration settings option specifies the path to PostgreSQL client certificate file.",
    )

    key_file: Optional[str] = Field(
        default=None,
        description="This configuration settings option specifies the path to PostgreSQL client key file.",
    )

    check_hostname: bool = Field(
        default=True,
        description="This configuration settings option specifies whether to verify PostgreSQL server hostname.",
    )


class SQLAlchemyConnectArgsSettings(PrefectBaseSettings):
    """
    Settings for controlling SQLAlchemy connection behavior; note that these settings only take effect when
    using a PostgreSQL database.
    """

    model_config: ClassVar[SettingsConfigDict] = build_settings_config(
        ("server", "database", "sqlalchemy", "connect_args")
    )

    application_name: Optional[str] = Field(
        default=None,
        description="Controls the application_name field for connections opened from the connection pool when using a PostgreSQL database with the Prefect backend.",
    )

    statement_cache_size: Optional[int] = Field(
        default=None,
        description="Controls statement cache size for PostgreSQL connections. Setting this to 0 is required when using PgBouncer in transaction mode. Defaults to None.",
    )

    prepared_statement_cache_size: Optional[int] = Field(
        default=None,
        description=(
            "Controls the size of the statement cache for PostgreSQL connections. "
            "When set to 0, statement caching is disabled. Defaults to None to use "
            "SQLAlchemy's default behavior."
        ),
    )

    tls: SQLAlchemyTLSSettings = Field(
        default_factory=SQLAlchemyTLSSettings,
        description="Settings for controlling SQLAlchemy mTLS behavior",
    )


class SQLAlchemySettings(PrefectBaseSettings):
    """
    Settings for controlling SQLAlchemy behavior; note that these settings only take effect when
    using a PostgreSQL database.
    """

    model_config: ClassVar[SettingsConfigDict] = build_settings_config(
        ("server", "database", "sqlalchemy")
    )

    connect_args: SQLAlchemyConnectArgsSettings = Field(
        default_factory=SQLAlchemyConnectArgsSettings,
        description="Settings for controlling SQLAlchemy connection behavior",
    )

    pool_size: int = Field(
        default=5,
        description="Controls connection pool size of database connection pools from the Prefect backend.",
        validation_alias=AliasChoices(
            AliasPath("pool_size"),
            "prefect_server_database_sqlalchemy_pool_size",
            "prefect_sqlalchemy_pool_size",
        ),
    )

    pool_recycle: int = Field(
        default=3600,
        description="This setting causes the pool to recycle connections after the given number of seconds has passed; set it to -1 to avoid recycling entirely.",
    )

    pool_timeout: Optional[float] = Field(
        default=30.0,
        description="Number of seconds to wait before giving up on getting a connection from the pool. Defaults to 30 seconds.",
    )

    max_overflow: int = Field(
        default=10,
        description="Controls maximum overflow of the connection pool. To prevent overflow, set to -1.",
        validation_alias=AliasChoices(
            AliasPath("max_overflow"),
            "prefect_server_database_sqlalchemy_max_overflow",
            "prefect_sqlalchemy_max_overflow",
        ),
    )


class ServerDatabaseSettings(PrefectBaseSettings):
    """
    Settings for controlling server database behavior
    """

    model_config: ClassVar[SettingsConfigDict] = build_settings_config(
        ("server", "database")
    )

    sqlalchemy: SQLAlchemySettings = Field(
        default_factory=SQLAlchemySettings,
        description="Settings for controlling SQLAlchemy behavior",
    )

    connection_url: Optional[SecretStr] = Field(
        default=None,
        description="""
        A database connection URL in a SQLAlchemy-compatible
        format. Prefect currently supports SQLite and Postgres. Note that all
        Prefect database engines must use an async driver - for SQLite, use
        `sqlite+aiosqlite` and for Postgres use `postgresql+asyncpg`.

        SQLite in-memory databases can be used by providing the url
        `sqlite+aiosqlite:///file::memory:?cache=shared&uri=true&check_same_thread=false`,
        which will allow the database to be accessed by multiple threads. Note
        that in-memory databases can not be accessed from multiple processes and
        should only be used for simple tests.
        """,
        validation_alias=AliasChoices(
            AliasPath("connection_url"),
            "prefect_server_database_connection_url",
            "prefect_api_database_connection_url",
        ),
    )

    driver: Optional[Literal["postgresql+asyncpg", "sqlite+aiosqlite"]] = Field(
        default=None,
        description=(
            "The database driver to use when connecting to the database. "
            "If not set, the driver will be inferred from the connection URL."
        ),
        validation_alias=AliasChoices(
            AliasPath("driver"),
            "prefect_server_database_driver",
            "prefect_api_database_driver",
        ),
    )

    host: Optional[str] = Field(
        default=None,
        description="The database server host.",
        validation_alias=AliasChoices(
            AliasPath("host"),
            "prefect_server_database_host",
            "prefect_api_database_host",
        ),
    )

    port: Optional[int] = Field(
        default=None,
        description="The database server port.",
        validation_alias=AliasChoices(
            AliasPath("port"),
            "prefect_server_database_port",
            "prefect_api_database_port",
        ),
    )

    user: Optional[str] = Field(
        default=None,
        description="The user to use when connecting to the database.",
        validation_alias=AliasChoices(
            AliasPath("user"),
            "prefect_server_database_user",
            "prefect_api_database_user",
        ),
    )

    name: Optional[str] = Field(
        default=None,
        description="The name of the Prefect database on the remote server, or the path to the database file for SQLite.",
        validation_alias=AliasChoices(
            AliasPath("name"),
            "prefect_server_database_name",
            "prefect_api_database_name",
        ),
    )

    password: Optional[SecretStr] = Field(
        default=None,
        description="The password to use when connecting to the database. Should be kept secret.",
        validation_alias=AliasChoices(
            AliasPath("password"),
            "prefect_server_database_password",
            "prefect_api_database_password",
        ),
    )

    echo: bool = Field(
        default=False,
        description="If `True`, SQLAlchemy will log all SQL issued to the database. Defaults to `False`.",
        validation_alias=AliasChoices(
            AliasPath("echo"),
            "prefect_server_database_echo",
            "prefect_api_database_echo",
        ),
    )

    migrate_on_start: bool = Field(
        default=True,
        description="If `True`, the database will be migrated on application startup.",
        validation_alias=AliasChoices(
            AliasPath("migrate_on_start"),
            "prefect_server_database_migrate_on_start",
            "prefect_api_database_migrate_on_start",
        ),
    )

    timeout: Optional[float] = Field(
        default=10.0,
        description="A statement timeout, in seconds, applied to all database interactions made by the Prefect backend. Defaults to 10 seconds.",
        validation_alias=AliasChoices(
            AliasPath("timeout"),
            "prefect_server_database_timeout",
            "prefect_api_database_timeout",
        ),
    )

    connection_timeout: Optional[float] = Field(
        default=5.0,
        description="A connection timeout, in seconds, applied to database connections. Defaults to `5`.",
        validation_alias=AliasChoices(
            AliasPath("connection_timeout"),
            "prefect_server_database_connection_timeout",
            "prefect_api_database_connection_timeout",
        ),
    )

    # handle deprecated fields

    def __getattribute__(self, name: str) -> Any:
        if name in ["sqlalchemy_pool_size", "sqlalchemy_max_overflow"]:
            warnings.warn(
                f"Setting {name} has been moved to the `sqlalchemy` settings group.",
                DeprecationWarning,
            )
            field_name = name.replace("sqlalchemy_", "")
            return getattr(super().__getattribute__("sqlalchemy"), field_name)
        return super().__getattribute__(name)

    # validators

    @model_validator(mode="before")
    @classmethod
    def set_deprecated_sqlalchemy_settings_on_child_model_and_warn(
        cls, values: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Set deprecated settings on the child model.
        """
        # Initialize sqlalchemy settings if not present
        if "sqlalchemy" not in values:
            values["sqlalchemy"] = SQLAlchemySettings()

        if "sqlalchemy_pool_size" in values:
            warnings.warn(
                "`sqlalchemy_pool_size` has been moved to the `sqlalchemy` settings group as `pool_size`.",
                DeprecationWarning,
            )
            if "pool_size" not in values["sqlalchemy"].model_fields_set:
                values["sqlalchemy"].pool_size = values["sqlalchemy_pool_size"]

        if "sqlalchemy_max_overflow" in values:
            warnings.warn(
                "`sqlalchemy_max_overflow` has been moved to the `sqlalchemy` settings group as `max_overflow`.",
                DeprecationWarning,
            )
            if "max_overflow" not in values["sqlalchemy"].model_fields_set:
                values["sqlalchemy"].max_overflow = values["sqlalchemy_max_overflow"]

        return values

    @model_validator(mode="after")
    def emit_warnings(self) -> Self:  # noqa: F821
        """More post-hoc validation of settings, including warnings for misconfigurations."""
        warn_on_database_password_value_without_usage(self)
        return self


def warn_on_database_password_value_without_usage(
    settings: ServerDatabaseSettings,
) -> None:
    """
    Validator for settings warning if the database password is set but not used.
    """
    db_password = (
        settings.password.get_secret_value()
        if isinstance(settings.password, SecretStr)
        else None
    )
    api_db_connection_url = (
        settings.connection_url.get_secret_value()
        if isinstance(settings.connection_url, SecretStr)
        else settings.connection_url
    )

    if (
        db_password
        and api_db_connection_url is not None
        and "PREFECT_API_DATABASE_PASSWORD" not in api_db_connection_url
        and "PREFECT_SERVER_DATABASE_PASSWORD" not in api_db_connection_url
        and db_password not in api_db_connection_url
        and quote_plus(db_password) not in api_db_connection_url
    ):
        warnings.warn(
            "PREFECT_SERVER_DATABASE_PASSWORD is set but not included in the "
            "PREFECT_SERVER_DATABASE_CONNECTION_URL. "
            "The provided password will be ignored."
        )
    return None



"""Adds block schema references and block document references

Revision ID: e73c6f1fe752
Revises: 33439667aeea
Create Date: 2022-05-28 08:16:50.141505

"""

import sqlalchemy as sa
from alembic import op

import prefect
from prefect.blocks.core import Block

# revision identifiers, used by Alembic.
revision = "e73c6f1fe752"
down_revision = "33439667aeea"
branch_labels = None
depends_on = None

# Used to update titles of existing storage block schemas. Need to update titles to match
# what the client generates to ensure the same checksum is generated on client and server.
BLOCK_SCHEMA_TITLE_MAP = {
    "S3 Storage": "S3StorageBlock",
    "Temporary Local Storage": "TempStorageBlock",
    "Local Storage": "LocalStorageBlock",
    "Google Cloud Storage": "GoogleCloudStorageBlock",
    "Azure Blob Storage": "AzureBlobStorageBlock",
    "KV Server Storage": "KVServerStorageBlock",
    "File Storage": "FileStorageBlock",
}


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "block_schema_reference",
        sa.Column(
            "id",
            prefect.server.utilities.database.UUID(),
            server_default=sa.text(
                "(\n    (\n        lower(hex(randomblob(4))) \n        || '-' \n       "
                " || lower(hex(randomblob(2))) \n        || '-4' \n        ||"
                " substr(lower(hex(randomblob(2))),2) \n        || '-' \n        ||"
                " substr('89ab',abs(random()) % 4 + 1, 1) \n        ||"
                " substr(lower(hex(randomblob(2))),2) \n        || '-' \n        ||"
                " lower(hex(randomblob(6)))\n    )\n    )"
            ),
            nullable=False,
        ),
        sa.Column(
            "created",
            prefect.server.utilities.database.Timestamp(timezone=True),
            server_default=sa.text("(strftime('%Y-%m-%d %H:%M:%f000', 'now'))"),
            nullable=False,
        ),
        sa.Column(
            "updated",
            prefect.server.utilities.database.Timestamp(timezone=True),
            server_default=sa.text("(strftime('%Y-%m-%d %H:%M:%f000', 'now'))"),
            nullable=False,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column(
            "parent_block_schema_id",
            prefect.server.utilities.database.UUID(),
            nullable=False,
        ),
        sa.Column(
            "reference_block_schema_id",
            prefect.server.utilities.database.UUID(),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["parent_block_schema_id"],
            ["block_schema.id"],
            name=op.f(
                "fk_block_schema_reference__parent_block_schema_id__block_schema"
            ),
            ondelete="cascade",
        ),
        sa.ForeignKeyConstraint(
            ["reference_block_schema_id"],
            ["block_schema.id"],
            name=op.f(
                "fk_block_schema_reference__reference_block_schema_id__block_schema"
            ),
            ondelete="cascade",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_block_schema_reference")),
    )
    op.create_index(
        op.f("ix_block_schema_reference__updated"),
        "block_schema_reference",
        ["updated"],
        unique=False,
    )
    op.create_table(
        "block_document_reference",
        sa.Column(
            "id",
            prefect.server.utilities.database.UUID(),
            server_default=sa.text(
                "(\n    (\n        lower(hex(randomblob(4))) \n        || '-' \n       "
                " || lower(hex(randomblob(2))) \n        || '-4' \n        ||"
                " substr(lower(hex(randomblob(2))),2) \n        || '-' \n        ||"
                " substr('89ab',abs(random()) % 4 + 1, 1) \n        ||"
                " substr(lower(hex(randomblob(2))),2) \n        || '-' \n        ||"
                " lower(hex(randomblob(6)))\n    )\n    )"
            ),
            nullable=False,
        ),
        sa.Column(
            "created",
            prefect.server.utilities.database.Timestamp(timezone=True),
            server_default=sa.text("(strftime('%Y-%m-%d %H:%M:%f000', 'now'))"),
            nullable=False,
        ),
        sa.Column(
            "updated",
            prefect.server.utilities.database.Timestamp(timezone=True),
            server_default=sa.text("(strftime('%Y-%m-%d %H:%M:%f000', 'now'))"),
            nullable=False,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column(
            "parent_block_document_id",
            prefect.server.utilities.database.UUID(),
            nullable=False,
        ),
        sa.Column(
            "reference_block_document_id",
            prefect.server.utilities.database.UUID(),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["parent_block_document_id"],
            ["block_document.id"],
            name=op.f(
                "fk_block_document_reference__parent_block_document_id__block_document"
            ),
            ondelete="cascade",
        ),
        sa.ForeignKeyConstraint(
            ["reference_block_document_id"],
            ["block_document.id"],
            name=op.f(
                "fk_block_document_reference__reference_block_document_id__block_document"
            ),
            ondelete="cascade",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_block_document_reference")),
    )
    op.create_index(
        op.f("ix_block_document_reference__updated"),
        "block_document_reference",
        ["updated"],
        unique=False,
    )

    # Update existing schemas to account for block schema references
    connection = op.get_bind()
    meta_data = sa.MetaData()
    meta_data.reflect(connection)
    BLOCK_SCHEMA = meta_data.tables["block_schema"]
    BLOCK_TYPE = meta_data.tables["block_type"]

    block_schemas = connection.execute(
        sa.select(
            BLOCK_SCHEMA.c.id, BLOCK_SCHEMA.c.fields, BLOCK_SCHEMA.c.block_type_id
        )
    )

    for id, fields, block_type_id in block_schemas:
        block_type_result = connection.execute(
            sa.select(BLOCK_TYPE.c.name).where(BLOCK_TYPE.c.id == block_type_id)
        ).first()
        block_type_name = block_type_result[0]
        updated_fields = {
            **fields,
            "block_type_name": block_type_name,
            "block_schema_references": {},
        }
        updated_title = BLOCK_SCHEMA_TITLE_MAP.get(block_type_name)
        if updated_title is not None:
            updated_fields["title"] = updated_title
        updated_checksum = Block._calculate_schema_checksum(updated_fields)
        connection.execute(
            sa.update(BLOCK_SCHEMA)
            .where(BLOCK_SCHEMA.c.id == id)
            .values(fields=updated_fields, checksum=updated_checksum)
        )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_block_document_reference__updated"),
        table_name="block_document_reference",
    )
    op.drop_table("block_document_reference")
    op.drop_index(
        op.f("ix_block_schema_reference__updated"), table_name="block_schema_reference"
    )
    op.drop_table("block_schema_reference")
    # ### end Alembic commands ###



from __future__ import annotations

import abc
import concurrent.futures
import threading
import uuid
import warnings
from collections.abc import Generator, Iterator
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Generic

from typing_extensions import NamedTuple, Self, TypeVar

from prefect._waiters import FlowRunWaiter
from prefect.client.orchestration import get_client
from prefect.exceptions import ObjectNotFound
from prefect.logging.loggers import get_logger
from prefect.states import Pending, State
from prefect.task_runs import TaskRunWaiter
from prefect.utilities.annotations import quote
from prefect.utilities.asyncutils import run_coro_as_sync
from prefect.utilities.collections import StopVisiting, visit_collection
from prefect.utilities.timeout import timeout as timeout_context

F = TypeVar("F")
R = TypeVar("R")

if TYPE_CHECKING:
    import logging

logger: "logging.Logger" = get_logger(__name__)


class PrefectFuture(abc.ABC, Generic[R]):
    """
    Abstract base class for Prefect futures. A Prefect future is a handle to the
    asynchronous execution of a run. It provides methods to wait for the
    to complete and to retrieve the result of the run.
    """

    def __init__(self, task_run_id: uuid.UUID):
        warnings.warn(
            "The __init__ method of PrefectFuture is deprecated and will be removed in a future release. "
            "If you are subclassing PrefectFuture, please implement the __init__ method in your subclass or "
            "subclass PrefectTaskRunFuture instead.",
            DeprecationWarning,
        )
        self._task_run_id = task_run_id
        self._final_state: State[R] | None = None

    @property
    def task_run_id(self) -> uuid.UUID:
        """The ID of the task run associated with this future"""
        warnings.warn(
            "The task_run_id property of PrefectFuture is deprecated and will be removed in a future release. "
            "If you are subclassing PrefectFuture, please implement the task_run_id property in your subclass or "
            "subclass PrefectTaskRunFuture instead.",
            DeprecationWarning,
        )

        return self._task_run_id

    @property
    def state(self) -> State:
        """The current state of the task run associated with this future"""
        warnings.warn(
            "The state property of PrefectFuture is deprecated and will be removed in a future release. "
            "If you are subclassing PrefectFuture, please implement the state property in your subclass or "
            "subclass PrefectTaskRunFuture instead.",
            DeprecationWarning,
        )

        if self._final_state:
            return self._final_state
        client = get_client(sync_client=True)
        try:
            task_run = client.read_task_run(task_run_id=self.task_run_id)
        except ObjectNotFound:
            # We'll be optimistic and assume this task will eventually start
            # TODO: Consider using task run events to wait for the task to start
            return Pending()
        return task_run.state or Pending()

    @abc.abstractmethod
    def wait(self, timeout: float | None = None) -> None:
        ...
        """
        Wait for the task run to complete.

        If the task run has already completed, this method will return immediately.

        Args:
            timeout: The maximum number of seconds to wait for the task run to complete.
              If the task run has not completed after the timeout has elapsed, this method will return.
        """

    @abc.abstractmethod
    def result(
        self,
        timeout: float | None = None,
        raise_on_failure: bool = True,
    ) -> R:
        ...
        """
        Get the result of the task run associated with this future.

        If the task run has not completed, this method will wait for the task run to complete.

        Args:
            timeout: The maximum number of seconds to wait for the task run to complete.
            If the task run has not completed after the timeout has elapsed, this method will return.
            raise_on_failure: If `True`, an exception will be raised if the task run fails.

        Returns:
            The result of the task run.
        """

    @abc.abstractmethod
    def add_done_callback(self, fn: Callable[["PrefectFuture[R]"], None]) -> None:
        """
        Add a callback to be run when the future completes or is cancelled.

        Args:
            fn: A callable that will be called with this future as its only argument when the future completes or is cancelled.
        """
        ...


class PrefectTaskRunFuture(PrefectFuture[R]):
    """
    A Prefect future that represents the eventual execution of a task run.
    """

    def __init__(self, task_run_id: uuid.UUID):
        self._task_run_id = task_run_id
        self._final_state: State[R] | None = None

    @property
    def task_run_id(self) -> uuid.UUID:
        """The ID of the task run associated with this future"""
        return self._task_run_id

    @property
    def state(self) -> State:
        """The current state of the task run associated with this future"""
        if self._final_state:
            return self._final_state
        client = get_client(sync_client=True)
        try:
            task_run = client.read_task_run(task_run_id=self.task_run_id)
        except ObjectNotFound:
            # We'll be optimistic and assume this task will eventually start
            # TODO: Consider using task run events to wait for the task to start
            return Pending()
        return task_run.state or Pending()


class PrefectWrappedFuture(PrefectTaskRunFuture[R], abc.ABC, Generic[R, F]):
    """
    A Prefect future that wraps another future object.

    Type Parameters:
        R: The return type of the future
        F: The type of the wrapped future
    """

    def __init__(self, task_run_id: uuid.UUID, wrapped_future: F):
        self._wrapped_future: F = wrapped_future
        super().__init__(task_run_id)

    @property
    def wrapped_future(self) -> F:
        """The underlying future object wrapped by this Prefect future"""
        return self._wrapped_future

    def add_done_callback(self, fn: Callable[[PrefectFuture[R]], None]) -> None:
        """Add a callback to be executed when the future completes."""
        if not self._final_state:

            def call_with_self(future: F):
                """Call the callback with self as the argument, this is necessary to ensure we remove the future from the pending set"""
                fn(self)

            self._wrapped_future.add_done_callback(call_with_self)
            return
        fn(self)


class PrefectConcurrentFuture(PrefectWrappedFuture[R, concurrent.futures.Future[R]]):
    """
    A Prefect future that wraps a concurrent.futures.Future. This future is used
    when the task run is submitted to a ThreadPoolExecutor.
    """

    def wait(self, timeout: float | None = None) -> None:
        try:
            result = self._wrapped_future.result(timeout=timeout)
        except concurrent.futures.TimeoutError:
            return
        if isinstance(result, State):
            self._final_state = result

    def result(
        self,
        timeout: float | None = None,
        raise_on_failure: bool = True,
    ) -> R:
        if not self._final_state:
            try:
                future_result = self._wrapped_future.result(timeout=timeout)
            except concurrent.futures.TimeoutError as exc:
                raise TimeoutError(
                    f"Task run {self.task_run_id} did not complete within {timeout} seconds"
                ) from exc

            if isinstance(future_result, State):
                self._final_state = future_result

            else:
                return future_result

        _result = self._final_state.result(
            raise_on_failure=raise_on_failure, _sync=True
        )
        return _result


class PrefectDistributedFuture(PrefectTaskRunFuture[R]):
    """
    Represents the result of a computation happening anywhere.

    This class is typically used to interact with the result of a task run
    scheduled to run in a Prefect task worker but can be used to interact with
    any task run scheduled in Prefect's API.
    """

    done_callbacks: list[Callable[[PrefectFuture[R]], None]] = []
    waiter = None

    def wait(self, timeout: float | None = None) -> None:
        return run_coro_as_sync(self.wait_async(timeout=timeout))

    async def wait_async(self, timeout: float | None = None) -> None:
        if self._final_state:
            logger.debug(
                "Final state already set for %s. Returning...", self.task_run_id
            )
            return

        # Ask for the instance of TaskRunWaiter _now_ so that it's already running and
        # can catch the completion event if it happens before we start listening for it.
        TaskRunWaiter.instance()

        # Read task run to see if it is still running
        async with get_client() as client:
            task_run = await client.read_task_run(task_run_id=self._task_run_id)
            if task_run.state is None:
                raise RuntimeError(
                    f"Task run {self.task_run_id} has no state which means it hasn't started yet."
                )
            if task_run.state.is_final():
                logger.debug(
                    "Task run %s already finished. Returning...",
                    self.task_run_id,
                )
                self._final_state = task_run.state
                return

            # If still running, wait for a completed event from the server
            logger.debug(
                "Waiting for completed event for task run %s...",
                self.task_run_id,
            )
            await TaskRunWaiter.wait_for_task_run(self._task_run_id, timeout=timeout)
            task_run = await client.read_task_run(task_run_id=self._task_run_id)
            if task_run.state.is_final():
                self._final_state = task_run.state
            return

    def result(
        self,
        timeout: float | None = None,
        raise_on_failure: bool = True,
    ) -> R:
        return run_coro_as_sync(
            self.result_async(timeout=timeout, raise_on_failure=raise_on_failure)
        )

    async def result_async(
        self,
        timeout: float | None = None,
        raise_on_failure: bool = True,
    ) -> R:
        if not self._final_state:
            await self.wait_async(timeout=timeout)
            if not self._final_state:
                raise TimeoutError(
                    f"Task run {self.task_run_id} did not complete within {timeout} seconds"
                )

        return await self._final_state.aresult(raise_on_failure=raise_on_failure)

    def add_done_callback(self, fn: Callable[[PrefectFuture[R]], None]) -> None:
        if self._final_state:
            fn(self)
            return
        TaskRunWaiter.instance()
        with get_client(sync_client=True) as client:
            task_run = client.read_task_run(task_run_id=self._task_run_id)
            if task_run.state.is_final():
                self._final_state = task_run.state
                fn(self)
                return
            TaskRunWaiter.add_done_callback(self._task_run_id, partial(fn, self))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, PrefectDistributedFuture):
            return False
        return self.task_run_id == other.task_run_id

    def __hash__(self) -> int:
        return hash(self.task_run_id)


class PrefectFlowRunFuture(PrefectFuture[R]):
    """
    A Prefect future that represents the eventual execution of a flow run.
    """

    def __init__(self, flow_run_id: uuid.UUID):
        self._flow_run_id = flow_run_id
        self._final_state: State[R] | None = None

    @property
    def flow_run_id(self) -> uuid.UUID:
        """The ID of the flow run associated with this future"""
        return self._flow_run_id

    @property
    def state(self) -> State:
        """The current state of the flow run associated with this future"""
        if self._final_state:
            return self._final_state
        client = get_client(sync_client=True)
        state = Pending()
        try:
            flow_run = client.read_flow_run(flow_run_id=self.flow_run_id)
            if flow_run.state:
                state = flow_run.state
        except ObjectNotFound:
            # We'll be optimistic and assume this flow run will eventually start
            pass
        return state

    def wait(self, timeout: float | None = None) -> None:
        return run_coro_as_sync(self.wait_async(timeout=timeout))

    async def wait_async(self, timeout: float | None = None) -> None:
        if self._final_state:
            logger.debug(
                "Final state already set for %s. Returning...", self.task_run_id
            )
            return

        # Ask for the instance of FlowRunWaiter _now_ so that it's already running and
        # can catch the completion event if it happens before we start listening for it.
        FlowRunWaiter.instance()

        # Read task run to see if it is still running
        async with get_client() as client:
            flow_run = await client.read_flow_run(flow_run_id=self._flow_run_id)
            if flow_run.state is None:
                raise RuntimeError(
                    f"Flow run {self.flow_run_id} has no state which means it hasn't started yet."
                )
            if flow_run.state and flow_run.state.is_final():
                logger.debug(
                    "Flow run %s already finished. Returning...",
                    self.flow_run_id,
                )
                self._final_state = flow_run.state
                return

            # If still running, wait for a completed event from the server
            logger.debug(
                "Waiting for completed event for flow run %s...",
                self.flow_run_id,
            )
            await FlowRunWaiter.wait_for_flow_run(self._flow_run_id, timeout=timeout)
            flow_run = await client.read_flow_run(flow_run_id=self._flow_run_id)
            if flow_run.state and flow_run.state.is_final():
                self._final_state = flow_run.state
            return

    def result(
        self,
        timeout: float | None = None,
        raise_on_failure: bool = True,
    ) -> R:
        return run_coro_as_sync(
            self.aresult(timeout=timeout, raise_on_failure=raise_on_failure)
        )

    async def aresult(
        self,
        timeout: float | None = None,
        raise_on_failure: bool = True,
    ) -> R:
        if not self._final_state:
            await self.wait_async(timeout=timeout)
            if not self._final_state:
                raise TimeoutError(
                    f"Task run {self.task_run_id} did not complete within {timeout} seconds"
                )

        return await self._final_state.aresult(raise_on_failure=raise_on_failure)

    def add_done_callback(self, fn: Callable[[PrefectFuture[R]], None]) -> None:
        if self._final_state:
            fn(self)
            return
        FlowRunWaiter.instance()
        with get_client(sync_client=True) as client:
            flow_run = client.read_flow_run(flow_run_id=self._flow_run_id)
            if flow_run.state and flow_run.state.is_final():
                self._final_state = flow_run.state
                fn(self)
                return
            FlowRunWaiter.add_done_callback(self._flow_run_id, partial(fn, self))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, PrefectFlowRunFuture):
            return False
        return self.flow_run_id == other.flow_run_id

    def __hash__(self) -> int:
        return hash(self.flow_run_id)


class PrefectFutureList(list[PrefectFuture[R]], Iterator[PrefectFuture[R]]):
    """
    A list of Prefect futures.

    This class provides methods to wait for all futures
    in the list to complete and to retrieve the results of all task runs.
    """

    def wait(self, timeout: float | None = None) -> None:
        """
        Wait for all futures in the list to complete.

        Args:
            timeout: The maximum number of seconds to wait for all futures to
                complete. This method will not raise if the timeout is reached.
        """
        wait(self, timeout=timeout)

    def result(
        self: Self,
        timeout: float | None = None,
        raise_on_failure: bool = True,
    ) -> list[R]:
        """
        Get the results of all task runs associated with the futures in the list.

        Args:
            timeout: The maximum number of seconds to wait for all futures to
                complete.
            raise_on_failure: If `True`, an exception will be raised if any task run fails.

        Returns:
            A list of results of the task runs.

        Raises:
            TimeoutError: If the timeout is reached before all futures complete.
        """
        try:
            with timeout_context(timeout):
                return [
                    future.result(raise_on_failure=raise_on_failure) for future in self
                ]
        except TimeoutError as exc:
            # timeout came from inside the task
            if "Scope timed out after {timeout} second(s)." not in str(exc):
                raise
            raise TimeoutError(
                f"Timed out waiting for all futures to complete within {timeout} seconds"
            ) from exc


def as_completed(
    futures: list[PrefectFuture[R]], timeout: float | None = None
) -> Generator[PrefectFuture[R], None]:
    unique_futures: set[PrefectFuture[R]] = set(futures)
    total_futures = len(unique_futures)
    pending = unique_futures
    try:
        with timeout_context(timeout):
            done = {f for f in unique_futures if f._final_state}  # type: ignore[privateUsage]
            pending = unique_futures - done
            yield from done

            finished_event = threading.Event()
            finished_lock = threading.Lock()
            finished_futures: list[PrefectFuture[R]] = []

            def add_to_done(future: PrefectFuture[R]):
                with finished_lock:
                    finished_futures.append(future)
                    finished_event.set()

            for future in pending:
                future.add_done_callback(add_to_done)

            while pending:
                finished_event.wait()
                with finished_lock:
                    done = finished_futures
                    finished_futures = []
                    finished_event.clear()

                for future in done:
                    pending.remove(future)
                    yield future

    except TimeoutError:
        raise TimeoutError(
            "%d (of %d) futures unfinished" % (len(pending), total_futures)
        )


class DoneAndNotDoneFutures(NamedTuple, Generic[R]):
    """A named 2-tuple of sets.

    multiple inheritance supported in 3.11+, use typing_extensions.NamedTuple
    """

    done: set[PrefectFuture[R]]
    not_done: set[PrefectFuture[R]]


def wait(
    futures: list[PrefectFuture[R]], timeout: float | None = None
) -> DoneAndNotDoneFutures[R]:
    """
    Wait for the futures in the given sequence to complete.

    Args:
        futures: The sequence of Futures to wait upon.
        timeout: The maximum number of seconds to wait. If None, then there
            is no limit on the wait time.

    Returns:
        A named 2-tuple of sets. The first set, named 'done', contains the
        futures that completed (is finished or cancelled) before the wait
        completed. The second set, named 'not_done', contains uncompleted
        futures. Duplicate futures given to *futures* are removed and will be
        returned only once.

    Examples:
        ```python
        @task
        def sleep_task(seconds):
            sleep(seconds)
            return 42

        @flow
        def flow():
            futures = random_task.map(range(10))
            done, not_done = wait(futures, timeout=5)
            print(f"Done: {len(done)}")
            print(f"Not Done: {len(not_done)}")
        ```
    """
    _futures = set(futures)
    done = {f for f in _futures if f._final_state}
    not_done = _futures - done
    if len(done) == len(_futures):
        return DoneAndNotDoneFutures(done, not_done)
    try:
        with timeout_context(timeout):
            for future in not_done.copy():
                future.wait()
                done.add(future)
                not_done.remove(future)
            return DoneAndNotDoneFutures(done, not_done)
    except TimeoutError:
        logger.debug("Timed out waiting for all futures to complete.")
        return DoneAndNotDoneFutures(done, not_done)


def resolve_futures_to_states(
    expr: PrefectFuture[R] | Any,
) -> PrefectFuture[R] | Any:
    """
    Given a Python built-in collection, recursively find `PrefectFutures` and build a
    new collection with the same structure with futures resolved to their final states.
    Resolving futures to their final states may wait for execution to complete.

    Unsupported object types will be returned without modification.
    """

    def _resolve_state(future: PrefectFuture[R]):
        future.wait()
        return future.state

    return _resolve_futures(
        expr,
        resolve_fn=_resolve_state,
    )


def resolve_futures_to_results(
    expr: PrefectFuture[R] | Any,
) -> Any:
    """
    Given a Python built-in collection, recursively find `PrefectFutures` and build a
    new collection with the same structure with futures resolved to their final results.
    Resolving futures to their final result may wait for execution to complete.

    Unsupported object types will be returned without modification.
    """

    def _resolve_result(future: PrefectFuture[R]) -> Any:
        future.wait()
        if future.state.is_completed():
            return future.result()
        else:
            raise Exception("At least one result did not complete successfully")

    return _resolve_futures(expr, resolve_fn=_resolve_result)


def _resolve_futures(
    expr: PrefectFuture[R] | Any,
    resolve_fn: Callable[[PrefectFuture[R]], Any],
) -> Any:
    """Helper function to resolve PrefectFutures in a collection."""
    futures: set[PrefectFuture[R]] = set()

    visit_collection(
        expr,
        visit_fn=partial(_collect_futures, futures),
        return_data=False,
        context={},
    )

    # If no futures were found, return the original expression
    if not futures:
        return expr

    # Resolve each future using the provided resolve function
    resolved_values = {future: resolve_fn(future) for future in futures}

    def replace_futures(expr: Any, context: Any) -> Any:
        # Expressions inside quotes should not be modified
        if isinstance(context.get("annotation"), quote):
            raise StopVisiting()

        if isinstance(expr, PrefectFuture):
            return resolved_values[expr]
        else:
            return expr

    return visit_collection(
        expr,
        visit_fn=replace_futures,
        return_data=True,
        context={},
    )


def _collect_futures(
    futures: set[PrefectFuture[R]], expr: Any | PrefectFuture[R], context: Any
) -> Any | PrefectFuture[R]:
    # Expressions inside quotes should not be traversed
    if isinstance(context.get("annotation"), quote):
        raise StopVisiting()

    if isinstance(expr, PrefectFuture):
        futures.add(expr)

    return expr



import prefect.server.services.cancellation_cleanup
import prefect.server.services.foreman
import prefect.server.services.late_runs
import prefect.server.services.pause_expirations
import prefect.server.services.scheduler
import prefect.server.services.telemetry



"""
The FailExpiredPauses service. Responsible for putting Paused flow runs in a Failed state if they are not resumed on time.
"""

import asyncio
from typing import Any, Optional

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession

import prefect.server.models as models
from prefect.server.database import PrefectDBInterface
from prefect.server.database.dependencies import db_injector
from prefect.server.database.orm_models import FlowRun
from prefect.server.schemas import states
from prefect.server.services.base import LoopService
from prefect.settings import PREFECT_API_SERVICES_PAUSE_EXPIRATIONS_LOOP_SECONDS
from prefect.settings.context import get_current_settings
from prefect.settings.models.server.services import ServicesBaseSetting
from prefect.types._datetime import now


class FailExpiredPauses(LoopService):
    """
    Fails flow runs that have been paused and never resumed
    """

    @classmethod
    def service_settings(cls) -> ServicesBaseSetting:
        return get_current_settings().server.services.pause_expirations

    def __init__(self, loop_seconds: Optional[float] = None, **kwargs: Any):
        super().__init__(
            loop_seconds=loop_seconds
            or PREFECT_API_SERVICES_PAUSE_EXPIRATIONS_LOOP_SECONDS.value(),
            **kwargs,
        )

        # query for this many runs to mark failed at once
        self.batch_size = 200

    @db_injector
    async def run_once(self, db: PrefectDBInterface) -> None:
        """
        Mark flow runs as failed by:

        - Querying for flow runs in a Paused state that have timed out
        - For any runs past the "expiration" threshold, setting the flow run state to a
          new `Failed` state
        """
        while True:
            async with db.session_context(begin_transaction=True) as session:
                query = (
                    sa.select(db.FlowRun)
                    .where(
                        db.FlowRun.state_type == states.StateType.PAUSED,
                    )
                    .limit(self.batch_size)
                )

                result = await session.execute(query)
                runs = result.scalars().all()

                # mark each run as failed
                for run in runs:
                    await self._mark_flow_run_as_failed(session=session, flow_run=run)

                # if no runs were found, exit the loop
                if len(runs) < self.batch_size:
                    break

        self.logger.info("Finished monitoring for late runs.")

    async def _mark_flow_run_as_failed(
        self, session: AsyncSession, flow_run: FlowRun
    ) -> None:
        """
        Mark a flow run as failed.

        Pass-through method for overrides.
        """
        if (
            flow_run.state is not None
            and flow_run.state.state_details.pause_timeout is not None
            and flow_run.state.state_details.pause_timeout < now("UTC")
        ):
            await models.flow_runs.set_flow_run_state(
                session=session,
                flow_run_id=flow_run.id,
                state=states.Failed(message="The flow was paused and never resumed."),
                force=True,
            )


if __name__ == "__main__":
    asyncio.run(FailExpiredPauses(handle_signals=True).start())



from __future__ import annotations

from functools import partial
from typing import Annotated, Any, Optional, TypeVar, Union
from typing_extensions import Literal
import orjson
import pydantic

from ._datetime import DateTime, Date
from .names import (
    Name,
    NameOrEmpty,
    NonEmptyishName,
    BANNED_CHARACTERS,
    WITHOUT_BANNED_CHARACTERS,
    MAX_VARIABLE_NAME_LENGTH,
)
from pydantic import (
    BeforeValidator,
    Field,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
    TypeAdapter,
)
from zoneinfo import available_timezones

T = TypeVar("T")

MAX_VARIABLE_VALUE_LENGTH = 5000

NonNegativeInteger = Annotated[int, Field(ge=0)]
PositiveInteger = Annotated[int, Field(gt=0)]
NonNegativeFloat = Annotated[float, Field(ge=0.0)]

TimeZone = Annotated[
    str,
    Field(
        default="UTC",
        pattern="|".join(
            [z for z in sorted(available_timezones()) if "localtime" not in z]
        ),
    ),
]


VariableValue = Union[
    StrictStr,
    StrictInt,
    StrictBool,
    StrictFloat,
    None,
    dict[str, Any],
    list[Any],
]


def check_variable_value(value: object) -> object:
    try:
        json_string = orjson.dumps(value)
    except orjson.JSONEncodeError:
        raise ValueError("Variable value must be serializable to JSON")

    if value is not None and len(json_string) > MAX_VARIABLE_VALUE_LENGTH:
        raise ValueError(
            f"Variable value must be less than {MAX_VARIABLE_VALUE_LENGTH} characters"
        )
    return value


StrictVariableValue = Annotated[VariableValue, BeforeValidator(check_variable_value)]

LaxUrl = Annotated[str, BeforeValidator(lambda x: str(x).strip())]

StatusCode = Annotated[int, Field(ge=100, le=599)]


def cast_none_to_empty_dict(value: Any) -> dict[str, Any]:
    if value is None:
        return {}
    return value


KeyValueLabels = Annotated[
    dict[str, Union[StrictBool, StrictInt, StrictFloat, str]],
    BeforeValidator(cast_none_to_empty_dict),
]


ListOfNonEmptyStrings = Annotated[
    list[str],
    BeforeValidator(lambda x: [str(s) for s in x if str(s).strip()]),
]


class SecretDict(pydantic.Secret[dict[str, Any]]):
    pass


def validate_set_T_from_delim_string(
    value: Union[str, T, set[T], None], type_: Any, delim: str | None = None
) -> set[T]:
    """
    "no-info" before validator useful in scooping env vars

    e.g. `PREFECT_CLIENT_RETRY_EXTRA_CODES=429,502,503` -> `{429, 502, 503}`
    e.g. `PREFECT_CLIENT_RETRY_EXTRA_CODES=429` -> `{429}`
    """
    if not value:
        return set()

    T_adapter = TypeAdapter(type_)
    delim = delim or ","
    if isinstance(value, str):
        return {T_adapter.validate_strings(s.strip()) for s in value.split(delim)}
    errors: list[pydantic.ValidationError] = []
    try:
        return {T_adapter.validate_python(value)}
    except pydantic.ValidationError as e:
        errors.append(e)
    try:
        return TypeAdapter(set[type_]).validate_python(value)
    except pydantic.ValidationError as e:
        errors.append(e)
    raise ValueError(f"Invalid set[{type_}]: {errors}")


ClientRetryExtraCodes = Annotated[
    Union[str, StatusCode, set[StatusCode], None],
    BeforeValidator(partial(validate_set_T_from_delim_string, type_=StatusCode)),
]


def parse_retry_delay_input(value: Any) -> Any:
    """
    Parses various inputs (string, int, float, list) into a format suitable
    for TaskRetryDelaySeconds (int, float, list[float], or None).
    Handles comma-separated strings for lists of delays.
    """
    if isinstance(value, str):
        stripped_value = value.strip()
        if not stripped_value:
            return None  # Treat empty or whitespace-only string as None

        delim = ","
        # Split and filter empty strings that result from multiple commas (e.g., "10,,20")
        parts = [s.strip() for s in stripped_value.split(delim) if s.strip()]

        if not parts:  # e.g., value was just "," or " , "
            return None

        def _parse_num_part(part_str: str) -> Union[float, int]:
            try:
                # Prefer float to align with list[float] in TaskRetryDelaySeconds
                return TypeAdapter(float).validate_strings(part_str)
            except pydantic.ValidationError:
                try:
                    return TypeAdapter(int).validate_strings(part_str)
                except pydantic.ValidationError as e_int:
                    raise ValueError(
                        f"Invalid number format '{part_str}' for retry delay."
                    ) from e_int

        if len(parts) == 1:
            return _parse_num_part(parts[0])
        else:
            return [_parse_num_part(p) for p in parts]

    # For non-string inputs (int, float, list, None, etc.), pass them through.
    # Pydantic will then validate them against Union[int, float, list[float], None].
    return value


TaskRetryDelaySeconds = Annotated[
    Union[str, int, float, list[float], None],
    BeforeValidator(parse_retry_delay_input),
]

LogLevel = Annotated[
    Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    BeforeValidator(lambda x: x.upper()),
]


def convert_none_to_empty_dict(v: Optional[KeyValueLabels]) -> KeyValueLabels:
    return v or {}


KeyValueLabelsField = Annotated[
    KeyValueLabels,
    Field(
        default_factory=dict,
        description="A dictionary of key-value labels. Values can be strings, numbers, or booleans.",
        examples=[{"key": "value1", "key2": 42}],
    ),
    BeforeValidator(convert_none_to_empty_dict),
]


__all__ = [
    "BANNED_CHARACTERS",
    "WITHOUT_BANNED_CHARACTERS",
    "ClientRetryExtraCodes",
    "Date",
    "DateTime",
    "LogLevel",
    "KeyValueLabelsField",
    "MAX_VARIABLE_NAME_LENGTH",
    "MAX_VARIABLE_VALUE_LENGTH",
    "NonNegativeInteger",
    "PositiveInteger",
    "ListOfNonEmptyStrings",
    "NonNegativeFloat",
    "Name",
    "NameOrEmpty",
    "NonEmptyishName",
    "SecretDict",
    "StatusCode",
    "StrictVariableValue",
    "TaskRetryDelaySeconds",
]



"""Add trgm index to block_document.name

Revision ID: 9c493c02ca6d
Revises: cef24af2ec34
Create Date: 2023-11-20 08:47:08.929487

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "9c493c02ca6d"
down_revision = "cef24af2ec34"
branch_labels = None
depends_on = None


def upgrade():
    with op.get_context().autocommit_block():
        op.execute(
            """
            CREATE INDEX CONCURRENTLY IF NOT EXISTS
            trgm_ix_block_document_name
            ON block_document USING gin (name gin_trgm_ops);
            """
        )


def downgrade():
    with op.get_context().autocommit_block():
        op.execute(
            """
            DROP INDEX CONCURRENTLY IF EXISTS
            trgm_ix_block_document_name;
            """
        )



"""json_variables

Revision ID: 2ac65f1758c2
Revises: 20fbd53b3cef
Create Date: 2024-05-21 12:29:43.948758

"""

import json

import sqlalchemy as sa
from alembic import op

from prefect.server.utilities.database import JSON

# revision identifiers, used by Alembic.
revision = "2ac65f1758c2"
down_revision = "20fbd53b3cef"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("variable", sa.Column("json_value", JSON, nullable=True))

    conn = op.get_bind()

    result = conn.execute(sa.text("SELECT id, value FROM variable"))
    rows = result.fetchall()

    for variable_id, value in rows:
        # these values need to be json compatible strings
        json_value = json.dumps(value)
        conn.execute(
            sa.text("UPDATE variable SET json_value = :json_value WHERE id = :id"),
            {"json_value": json_value, "id": variable_id},
        )

    with op.batch_alter_table("variable") as batch_op:
        batch_op.drop_column("value")
        batch_op.alter_column("json_value", new_column_name="value")


def downgrade():
    op.add_column("variable", sa.Column("string_value", sa.String, nullable=True))

    conn = op.get_bind()

    result = conn.execute(sa.text("SELECT id, value FROM variable"))
    rows = result.fetchall()

    for variable_id, value in rows:
        string_value = json.loads(str(value))
        conn.execute(
            sa.text("UPDATE variable SET string_value = :string_value WHERE id = :id"),
            {"string_value": string_value, "id": variable_id},
        )

    with op.batch_alter_table("variable") as batch_op:
        batch_op.drop_column("value")
        batch_op.alter_column("string_value", new_column_name="value", nullable=False)



"""Rename Block to BlockDocument and BlockSpec to BlockSchema

Revision ID: d38c5e6a9115
Revises: 14dc68cc5853
Create Date: 2022-04-19 13:28:03.661481

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "d38c5e6a9115"
down_revision = "14dc68cc5853"
branch_labels = None
depends_on = None


def upgrade():
    op.rename_table("block_spec", "block_schema")
    op.rename_table("block", "block_document")

    with op.batch_alter_table("block_document", schema=None) as batch_op:
        batch_op.drop_index("ix_block__is_default_storage_block")
        batch_op.drop_index("ix_block__name")
        batch_op.drop_index("ix_block__updated")
        batch_op.drop_index("uq_block__spec_id_name")
        batch_op.alter_column("block_spec_id", new_column_name="block_schema_id")
        batch_op.alter_column(
            "is_default_storage_block",
            new_column_name="is_default_storage_block_document",
        )
        batch_op.drop_constraint("fk_block__block_spec_id__block_spec")
        batch_op.drop_constraint("pk_block_data")

    with op.batch_alter_table("block_document", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_block_document__is_default_storage_block_document"),
            ["is_default_storage_block_document"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_block_document__name"), ["name"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_block_document__updated"), ["updated"], unique=False
        )
        batch_op.create_index(
            "uq_block__schema_id_name", ["block_schema_id", "name"], unique=True
        )

    with op.batch_alter_table("block_schema", schema=None) as batch_op:
        batch_op.drop_index("ix_block_spec__type")
        batch_op.drop_index("ix_block_spec__updated")
        batch_op.drop_index("uq_block_spec__name_version")
        batch_op.create_index("ix_block_schema__type", ["type"], unique=False)
        batch_op.create_index(
            batch_op.f("ix_block_schema__updated"), ["updated"], unique=False
        )
        batch_op.create_index(
            "uq_block_schema__name_version", ["name", "version"], unique=True
        )
        batch_op.drop_constraint("pk_block_spec")
        batch_op.create_primary_key("pk_block_schema", ["id"])

    with op.batch_alter_table("block_document", schema=None) as batch_op:
        batch_op.create_primary_key("pk_block_document", ["id"])
        batch_op.create_foreign_key(
            batch_op.f("fk_block__block_schema_id__block_schema"),
            "block_schema",
            ["block_schema_id"],
            ["id"],
            ondelete="cascade",
        )


def downgrade():
    op.rename_table("block_schema", "block_spec")
    op.rename_table("block_document", "block")

    with op.batch_alter_table("block", schema=None) as batch_op:
        batch_op.drop_index("ix_block_document__is_default_storage_block_document")
        batch_op.drop_index("ix_block_document__name")
        batch_op.drop_index("ix_block_document__updated")
        batch_op.drop_index("uq_block__schema_id_name")
        batch_op.drop_constraint("fk_block__block_schema_id__block_schema")
        batch_op.drop_constraint("pk_block_document")

    with op.batch_alter_table("block", schema=None) as batch_op:
        batch_op.alter_column("block_schema_id", new_column_name="block_spec_id")
        batch_op.alter_column(
            "is_default_storage_block_document",
            new_column_name="is_default_storage_block",
        )

    with op.batch_alter_table("block", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_block__is_default_storage_block"),
            ["is_default_storage_block"],
            unique=False,
        )
        batch_op.create_index(batch_op.f("ix_block__name"), ["name"], unique=False)
        batch_op.create_index(
            batch_op.f("ix_block__updated"), ["updated"], unique=False
        )
        batch_op.create_index(
            "uq_block__spec_id_name", ["block_spec_id", "name"], unique=True
        )

    with op.batch_alter_table("block_spec", schema=None) as batch_op:
        batch_op.drop_index("ix_block_schema__type")
        batch_op.drop_index("ix_block_schema__updated")
        batch_op.drop_index("uq_block_schema__name_version")
        batch_op.create_index("ix_block_spec__type", ["type"], unique=False)
        batch_op.create_index(
            batch_op.f("ix_block_spec__updated"), ["updated"], unique=False
        )
        batch_op.create_index(
            "uq_block_spec__name_version", ["name", "version"], unique=True
        )
        batch_op.drop_constraint("pk_block_schema")
        batch_op.create_primary_key("pk_block_spec", ["id"])

    with op.batch_alter_table("block", schema=None) as batch_op:
        batch_op.create_primary_key("pk_block_data", ["id"])
        batch_op.create_foreign_key(
            batch_op.f("fk_block__block_spec_id__block_spec"),
            "block_spec",
            ["block_spec_id"],
            ["id"],
            ondelete="cascade",
        )



"""Worker status field

Revision ID: a8e62d4c72cf
Revises: 75c8f17b8b51
Create Date: 2024-04-25 15:51:20.234353

"""

import sqlalchemy as sa
from alembic import op

from prefect.server.utilities.database import UUID, Timestamp

# revision identifiers, used by Alembic.
revision = "a8e62d4c72cf"
down_revision = "75c8f17b8b51"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("work_pool", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "last_transitioned_status_at", Timestamp(timezone=True), nullable=True
            ),
        )
        batch_op.add_column(sa.Column("last_status_event_id", UUID(), nullable=True))

    with op.batch_alter_table("worker", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "status",
                sa.Enum("ONLINE", "OFFLINE", name="worker_status"),
                server_default="OFFLINE",
                nullable=False,
            )
        )


def downgrade():
    with op.batch_alter_table("worker", schema=None) as batch_op:
        batch_op.drop_column("status")

    with op.batch_alter_table("work_pool", schema=None) as batch_op:
        batch_op.drop_column("last_transitioned_status_at")
        batch_op.drop_column("last_status_event_id")



from contextvars import ContextVar
from typing import Any, ClassVar

from typing_extensions import Self

from prefect.client.orchestration import get_client
from prefect.context import ContextModel, Field


class ConcurrencyContext(ContextModel):
    __var__: ClassVar[ContextVar[Self]] = ContextVar("concurrency")

    # Track the slots that have been acquired but were not able to be released
    # due to cancellation or some other error. These slots are released when
    # the context manager exits.
    cleanup_slots: list[tuple[list[str], int, float]] = Field(
        default_factory=lambda: []
    )

    def __exit__(self, *exc_info: Any) -> None:
        if self.cleanup_slots:
            with get_client(sync_client=True) as client:
                for names, occupy, occupancy_seconds in self.cleanup_slots:
                    client.release_concurrency_slots(
                        names=names, slots=occupy, occupancy_seconds=occupancy_seconds
                    )

        return super().__exit__(*exc_info)
